from __future__ import annotations

import re
from dataclasses import dataclass, field
from typing import Mapping, Optional, Sequence

from gp.distantsupervision.make_dataset.utils import FilterMixin, StrPath
from kgdata.models.entity import Entity
from kgdata.models.ont_class import OntologyClass
from kgdata.wikidata.models.wdclass import WDClass
from kgdata.wikidata.models.wdentity import WDEntity
from sm.dataset import FullTable

# fmt: off
DEFAULT_IGNORE_ENTS = {
    "Q13406463"  # wikimedia list article
}
# fmt: on


@dataclass
class FilterByEntTypeArgs:
    ignore_types: set[str] = field(
        default_factory=lambda: DEFAULT_IGNORE_ENTS,
        metadata={"help": "Entities belong to these types will be ignored"},
    )
    noisy_threshold: float = 0.5


class FilterByEntType(FilterMixin):
    VERSION = 100

    def __init__(
        self,
        args: FilterByEntTypeArgs,
        entities: Mapping[str, WDEntity | Entity],
        classes: Mapping[str, OntologyClass],
        log_file: Optional[StrPath] = None,
    ):
        super().__init__(log_file)
        self.args = args
        self.entities = entities
        self.classes = classes

    def is_noisy(self, table: FullTable, ci: int) -> tuple[bool, str]:
        nrows, ncols = table.table.shape()

        n_noisy_cells = 0
        for ri in range(nrows):
            if len(table.links[ri, ci]) == 0:
                continue
            entids = {
                str(entid) for link in table.links[ri, ci] for entid in link.entities
            }
            if any(self.has_ignore_type(eid) for eid in entids):
                n_noisy_cells += 1

        noisy_percentage = n_noisy_cells / nrows
        return (
            noisy_percentage >= self.args.noisy_threshold,
            f"{noisy_percentage:.03f} cells have ignore type",
        )

    def has_ignore_type(self, entid: str):
        ignore_types = self.args.ignore_types
        for cid in self.entities[entid].instance_of():
            if cid in ignore_types or not ignore_types.isdisjoint(
                self.classes[cid].ancestors
            ):
                return True
        return False



from __future__ import annotations

from copy import copy

import numpy as np
from gp.actors.data.prelude import GPExample
from gp.entity_linking.cangen.common import CanEnt, TableCanGenResult
from gp.entity_linking.cangen.oracle_model import CanGenOracleMethod
from gp.entity_linking.candidate_ranking.feats.feat_fn import CRFeatFn, CRFeatFnArgs
from nptyping import Bool, Float64, Int32, NDArray, Object, Shape
from ream.cache_helper import Cache, CacheableFn
from ream.data_model_helper._numpy_model import NumpyDataModel


class CRTableCan(NumpyDataModel):
    __slots__ = [
        "index",
        "cell",
        "col_index",
        "row_index",
        "ent_id",
        "ent_retrieve_score",
        "is_correct",
    ]
    # mapping from col_index => row_index => (start, end)
    index: dict[int, dict[int, tuple[int, int]]]
    cell: NDArray[Shape["*"], Object]  # content of the cell
    col_index: NDArray[Shape["*"], Int32]  # column index of the cell
    row_index: NDArray[Shape["*"], Int32]  # row index of the cell
    ent_id: NDArray[Shape["*"], Object]
    ent_retrieve_score: NDArray[Shape["*"], Float64]  # retrieval score of the entity
    is_correct: NDArray[Shape["*"], Bool]  # whether the cell is correct

    def to_table_cangen_result(self) -> TableCanGenResult:
        return TableCanGenResult(
            index=self.index,
            ent_id=self.ent_id,
            ent_score=self.ent_retrieve_score,
        )


CRTableCan.init()


class GetCandidatesFn(CRFeatFn):
    use_args = [CRFeatFnArgs.add_missing_gold, CRFeatFnArgs.remove_nil_entity]

    @Cache.cache(
        backend=Cache.sqlite.serde(
            cls=CRTableCan,
            filename=lambda slf, ex: slf.cache_file + "/data.sqlite",
            mem_persist=True,
        ),
        cache_key=lambda self, ex: ex.id,
        disable="disable",
    )
    def __call__(self, ex: GPExample) -> CRTableCan:
        cans = self.cangen_actor(ex)
        gold_cans = CanGenOracleMethod().get_candidates(
            [ex], [list(range(ex.table.ncols()))]
        )[0]

        if self.args.add_missing_gold:
            # add missing gold entities
            shp = ex.table.table.shape()
            mcan = cans.to_matrix(shp)
            mgoldcan = gold_cans.to_matrix(shp)

            new_mcan = mcan.map_with_index(
                lambda ri, ci, cans: GetCandidatesFn.merge_gold_cans(
                    cans, mgoldcan[ri][ci], self.args.remove_nil_entity
                )
            )
            cans = TableCanGenResult.from_matrix(new_mcan)
        else:
            assert (
                not self.args.remove_nil_entity
            ), "remove_nil_entity is only valid when add_missing_gold is True"

        cell = []
        col_index = []
        row_index = []
        is_correct = []

        for ci in cans.index:
            for ri, (start, end) in cans.index[ci].items():
                text = ex.table.table[ri, ci]

                cell.extend((text for _ in range(start, end)))
                col_index.extend((ci for _ in range(start, end)))
                row_index.extend((ri for _ in range(start, end)))

                gold_ids = {x.id for x in gold_cans.get_cell_candidates(ri, ci)}
                for j in range(start, end):
                    is_correct.append(cans.ent_id[j] in gold_ids)

        return CRTableCan(
            index=cans.index,
            cell=np.asarray(cell, dtype=np.object_),
            col_index=np.asarray(col_index, dtype=np.int64),
            row_index=np.asarray(row_index, dtype=np.int64),
            ent_id=cans.ent_id,
            ent_retrieve_score=cans.ent_score,
            is_correct=np.asarray(is_correct, dtype=np.bool_),
        )

    @staticmethod
    def merge_gold_cans(
        cans: list[CanEnt], gold_cans: list[CanEnt], remove_nil_entity: bool
    ):
        # remove candidates with NIL entity -- i.e., cells without gold entity
        if len(gold_cans) == 0 and remove_nil_entity:
            return []

        can_ids = {can.id for can in cans}
        gold_cans = [can for can in gold_cans if can.id not in can_ids]

        if len(gold_cans) > 0:
            # although `to_matrix` function create a new copy, this function still
            # copy because we cannot guarantee that usages in which the cans is already a temp copy
            cans = copy(cans)

            for gold_can in gold_cans:
                cans.append(gold_can)

        return cans



from __future__ import annotations

from typing import Sequence

from gp.entity_linking.candidate_generation.common import (
    CanEnt,
    CanGenMethod,
    TableCanGenResult,
)
from sm.dataset import Example, FullTable
from sm.inputs.prelude import Link


class CanGenOracleMethod(CanGenMethod):
    VERSION = 100

    def __init__(self, impute_score: float = 1.0):
        self.impute_score = impute_score

    def get_candidates(
        self,
        examples: Sequence[Example[FullTable]],
        entity_columns: Sequence[Sequence[int]],
    ) -> list[TableCanGenResult]:
        output = []

        for ei, ex in enumerate(examples):
            exentcols = set(entity_columns[ei])
            output.append(
                TableCanGenResult.from_matrix(
                    ex.table.links.map_with_index(
                        lambda ri, ci, links: (
                            map_links(links, self.impute_score)
                            if ci in exentcols
                            else []
                        )
                    )
                )
            )

        return output


def map_links(links: list[Link], impute_score: float) -> list[CanEnt]:
    gold_ents = {}
    for link in links:
        gold_ents.update((str(entid), impute_score) for entid in link.entities)
    return [CanEnt(entid, score) for entid, score in gold_ents.items()]



from __future__ import annotations

import gramsplus.core as gcore
import numpy as np
from gp.actors.data.prelude import GPExample
from gp.entity_linking.candidate_ranking.feats.feat_fn import CRFeatFn, CRFeatFnArgs
from gp.entity_linking.candidate_ranking.feats.get_candidates import CRTableCan, GetCandidatesFn
from gp.misc.conversions import to_rust_table
from gramsplus.semanticmodeling.text_parser import TextParser
from nptyping import Single
from ream.cache_helper import Cache
from ream.data_model_helper._numpy_model import SingleNumpyArray


class GetFreqMatchDataFn(CRFeatFn):
    use_args = [
        CRFeatFnArgs.text_parser_cfg,
        CRFeatFnArgs.literal_matcher_cfg,
    ]

    get_candidates: GetCandidatesFn

    REASONABLY_LARGE_NO_COLUMN = 20

    def set_args(self, args: CRFeatFnArgs):
        super().set_args(args)
        self.text_parser = TextParser(args.text_parser_cfg)
        self.literal_matcher = gcore.literal_matchers.LiteralMatcher(
            args.literal_matcher_cfg.to_rust()
        )
        return self

    @Cache.cache(
        backend=Cache.sqlite.serde(
            cls=SingleNumpyArray,
            filename=lambda slf, ex: slf.cache_file + "/data.sqlite",
            mem_persist=True,
        ),
        cache_key=lambda self, ex: ex.id,
        disable="disable",
    )
    def __call__(self, ex: GPExample):
        """Calculate how many incoming and outgoing relationships to other cells in a row of a candidate entity"""
        nrows, ncols = ex.table.table.shape()

        cans = self.get_candidates(ex)
        table = to_rust_table(ex, cans.to_table_cangen_result())
        table_cells = gcore.models.TableCells(
            [
                [
                    self.text_parser.parse(ex.table.table[ri, ci]).to_rust()
                    for ci in range(ncols)
                ]
                for ri in range(nrows)
            ]
        )

        table_res = gcore.algorithms.extract_candidate_entity_link_freqs(
            self.store.kgdb_service(ex.kgname).kgdb.rudb,
            table,
            table_cells,
            None,
            self.literal_matcher,
            ignored_columns=[],
            ignored_props=[],
            allow_same_ent_search=False,
            allow_ent_matching=True,
            use_context=True,
            deterministic_order=False,
            parallel=True,
        )
        freqs = GetFreqMatchDataFn.normalize_results(
            ex,
            cans,
            table_res,
        )
        return SingleNumpyArray(np.asarray(freqs) / self.REASONABLY_LARGE_NO_COLUMN)

    @Cache.flat_cache(
        backend=Cache.sqlite.serde(
            cls=SingleNumpyArray,
            filename=lambda slf, ex: slf.cache_file + "/data.sqlite",
            mem_persist=True,
        ),
        cache_key=lambda self, ex, verbose=False: ex.id,
        disable="disable",
    )
    def batch_call(self, exs: list[GPExample], verbose: bool = False):
        if len(exs) == 0:
            return SingleNumpyArray(np.array([]))

        in_tables = []
        in_table_cells = []

        ex_cans = self.get_candidates.batch_call(exs)

        for ei, ex in enumerate(exs):
            nrows, ncols = ex.table.table.shape()
            in_tables.append(to_rust_table(ex, ex_cans[ei].to_table_cangen_result()))
            in_table_cells.append(
                gcore.models.TableCells(
                    [
                        [
                            self.text_parser.parse(ex.table.table[ri, ci]).to_rust()
                            for ci in range(ncols)
                        ]
                        for ri in range(nrows)
                    ]
                )
            )

        table_ress = gcore.algorithms.par_extract_candidate_entity_link_freqs(
            self.store.kgdb_service(exs[0].kgname).kgdb.rudb,
            in_tables,
            in_table_cells,
            None,
            self.literal_matcher,
            ignored_columns=[[] for _ in range(len(in_tables))],
            ignored_props=[],
            allow_same_ent_search=False,
            allow_ent_matching=True,
            use_context=True,
            deterministic_order=False,
            verbose=verbose,
        )
        return [
            SingleNumpyArray(
                GetFreqMatchDataFn.normalize_results(
                    ex,
                    cans,
                    res,
                )
                / self.REASONABLY_LARGE_NO_COLUMN
            )
            for ex, cans, res in zip(exs, ex_cans, table_ress)
        ]

    @staticmethod
    def normalize_results(
        example: GPExample,
        crcan_base: CRTableCan,
        link_freqs: list[dict[str, list[int]]],
    ):
        nrows, ncols = example.table.table.shape()
        row_index = crcan_base.row_index
        col_index = crcan_base.col_index
        matched_freqs = []
        for i in range(len(crcan_base)):
            ri = row_index[i]
            ci = col_index[i]
            canid = crcan_base.ent_id[i]

            freq = link_freqs[ri * ncols + ci].get(canid, 0)
            matched_freqs.append(freq)
        return np.asarray(matched_freqs)



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Literal, Optional

import numpy as np
import serde.json
import serde.pickle
import torch
import torch.nn as nn
import torch.nn.functional as F
from gp.actors.data.prelude import GPExample
from gp.entity_linking.candidate_ranking.common import (
    BaseTorchCanRankModel,
    BaseTorchCanRankOutput,
)
from gp.entity_linking.candidate_ranking.feats.make_cr_dataset import (
    CRDatasetBuilder,
    make_pairwise_v3_nocontext_dataset,
)
from sm.misc.funcs import assert_not_null
from torchmetrics.retrieval import RetrievalHitRate, RetrievalMRR


@dataclass
class PairwiseCanRankV3Args:
    base_feat_dim: (
        int  # dimension of the heuristic features (e.g., string similarity, pagerank)
    )
    embed_dim: int  # dimension of the sentence embedding
    pgrnk_stat_file: (
        Path | str
    )  # path to the pagerank statistics (e.g., pagerank_en.pkl)
    pgrnk_norm: Literal["normalization", "none", "standardization"] = (
        "normalization"  # normalization method, normalization seems to be the best
    )
    training_topk: int = 20  # top-k candidates to use for training
    training_add_missing_gold: bool = (
        True  # whether to add the gold candidate to the training set
    )


class PairwiseCanRankV3(BaseTorchCanRankModel[PairwiseCanRankV3Args]):
    VERSION = 100
    EVAL_BATCH_SIZE = 10000
    EXPECTED_EVAL_ARGS = {"features", "entity_desc_embedding", "header_embedding"}
    EXPECTED_ARGS = EXPECTED_EVAL_ARGS.union({"label"})

    def __init__(
        self,
        base_feat_dim: int,
        embed_dim: int,
        log_pagerank: Optional[tuple[float, float]] = None,
    ):
        super().__init__()

        self.log_pagerank = log_pagerank
        hidden_dim = 3 * base_feat_dim
        self.embed_dim = embed_dim

        self.context_mat_diag = nn.Parameter(torch.ones(self.embed_dim))
        # two hidden layers
        self.cmp = nn.Sequential(
            nn.Linear(base_feat_dim + 1, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, 2),
        )
        self.class_weights = nn.Parameter(
            torch.FloatTensor([1, 1]), requires_grad=False
        )
        self.args: Optional[PairwiseCanRankV3Args] = None
        self.save_hyperparameters()

        self.metrics = {}
        for subset in ["train", "val", "test"]:
            # note: skip the NIL or dangling entity as MRR & Hit@k aren't designed to measure NIL/dangling -- use precision/recall instead.
            setattr(self, f"{subset}_mrr", RetrievalMRR(empty_target_action="skip"))
            for k in [1, 3, 5, 10]:
                setattr(
                    self,
                    f"{subset}_hit@{k}",
                    RetrievalHitRate(empty_target_action="skip", top_k=k),
                )

    @classmethod
    def from_args(cls, args: PairwiseCanRankV3Args):
        pgrnk_stat = serde.pickle.deser(args.pgrnk_stat_file)
        if args.pgrnk_norm == "normalization":
            log_pagerank = (
                float(np.log(pgrnk_stat["min"] + 1e-20)),
                float(np.log(pgrnk_stat["max"]) - np.log(pgrnk_stat["min"] + 1e-20)),
            )
        elif args.pgrnk_norm == "standardization":
            pgrnk_mean = pgrnk_stat["meanlog"]
            pgrnk_std = pgrnk_stat["stdlog"]
            log_pagerank = (pgrnk_mean, pgrnk_std)
        else:
            assert args.pgrnk_norm != "none"
            log_pagerank = None

        model = cls(
            base_feat_dim=args.base_feat_dim,
            embed_dim=args.embed_dim,
            log_pagerank=log_pagerank,
        )
        model.args = args
        return model

    def forward(
        self, features, entity_desc_embedding, header_embedding, label=None
    ) -> BaseTorchCanRankOutput:
        if self.embed_dim == 0:
            ctx_score = torch.zeros((features.shape[0], 1)).to(features.device)
        else:
            ctx_score = (
                entity_desc_embedding
                * self.context_mat_diag.view(1, -1)
                * header_embedding
            ).sum(dim=1, keepdim=True)

        if self.log_pagerank is not None:
            norm_pg_rnk = (
                torch.log(features[:, -1:] + 1e-20) - self.log_pagerank[0]
            ) / self.log_pagerank[1]
            inputs = torch.cat([features[:, :-1], norm_pg_rnk, ctx_score], dim=1)
        else:
            inputs = torch.cat([features, ctx_score], dim=1)

        logits = self.cmp(inputs)

        if label is not None:
            loss = F.nll_loss(
                input=F.log_softmax(logits, dim=1),
                target=label,
                weight=self.class_weights,
            )
        else:
            loss = None

        return BaseTorchCanRankOutput(loss=loss, probs=F.softmax(logits, dim=1)[:, 1])

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=1e-4)
        return optimizer

    def training_step(self, batch, batch_idx):
        output = self._pl_step(batch, batch_idx, "train")
        return output.loss

    def validation_step(self, batch, batch_idx):
        self._pl_step(batch, batch_idx, "val")

    def test_step(self, batch, batch_idx):
        self._pl_step(batch, batch_idx, "test")

    def _pl_step(self, batch, batch_idx, prefix):
        batch_size = batch["label"].shape[0]
        output = self.forward(**{arg: batch[arg] for arg in self.EXPECTED_ARGS})

        self.log(
            f"{prefix}_loss",
            assert_not_null(output.loss),
            prog_bar=True,
            batch_size=batch_size,
        )
        if prefix != "train":
            # see usage here: https://torchmetrics.readthedocs.io/en/stable/pages/lightning.html
            preds, target, indexes = (
                output.probs,
                batch["label"],
                batch["cell_id"].long(),
            )
            # although the default of on_step and on_epoch for evaluation/test steps are correctly set by pytorch-lightning,
            # we specify them explicitly here for the training step. we need to do this as retrieval metrics are only intended to be used globally
            # see more: https://torchmetrics.readthedocs.io/en/v0.8.2/pages/retrieval.html and https://lightning.ai/docs/pytorch/stable/visualize/logging_advanced.html
            getattr(self, f"{prefix}_mrr")(preds, target, indexes)
            getattr(self, f"{prefix}_hit@1")(preds, target, indexes)
            getattr(self, f"{prefix}_hit@3")(preds, target, indexes)
            getattr(self, f"{prefix}_hit@5")(preds, target, indexes)
            getattr(self, f"{prefix}_hit@10")(preds, target, indexes)
            self.log(
                f"{prefix}_mrr",
                getattr(self, f"{prefix}_mrr"),
                prog_bar=True,
                on_epoch=True,
                on_step=False,
                batch_size=batch_size,
            )
            self.log(
                f"{prefix}_hit@1",
                getattr(self, f"{prefix}_hit@1"),
                prog_bar=True,
                on_epoch=True,
                on_step=False,
                batch_size=batch_size,
            )
            self.log(
                f"{prefix}_hit@3",
                getattr(self, f"{prefix}_hit@3"),
                on_epoch=True,
                on_step=False,
                batch_size=batch_size,
            )
            self.log(
                f"{prefix}_hit@5",
                getattr(self, f"{prefix}_hit@5"),
                on_epoch=True,
                on_step=False,
                batch_size=batch_size,
            )
            self.log(
                f"{prefix}_hit@10",
                getattr(self, f"{prefix}_hit@10"),
                on_epoch=True,
                on_step=False,
                batch_size=batch_size,
            )
        return output

    def make_dataset(
        self,
        store: CRDatasetBuilder,
        examples: list[GPExample],
        verbose: bool = False,
    ):
        return make_pairwise_v3_nocontext_dataset(store, examples, verbose)



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, Optional, Sequence, TypeVar

import lightning.pytorch as pl
import torch
from gp.actors.data import GPExample
from gp.entity_linking.candidate_ranking.feats.make_cr_dataset import CRDatasetBuilder
from gp.misc.dataset import ColumnarDataset
from smml.data_model_helper import SingleNumpyArray
from torch.utils.data import DataLoader
from tqdm.auto import tqdm

TableCanGenUpdateScores = SingleNumpyArray


class CanRankMethod:

    @abstractmethod
    def batch_rank_candidates(
        self,
        examples: Sequence[GPExample],
        verbose: bool = False,
    ) -> list[TableCanGenUpdateScores]:
        pass


C = TypeVar("C")
A = TypeVar("A")


@dataclass
class BaseTorchCanRankOutput:
    loss: Optional[torch.Tensor]
    probs: torch.Tensor


class BaseTorchCanRankModel(ABC, Generic[A], pl.LightningModule):
    EVAL_BATCH_SIZE = 100
    EXPECTED_EVAL_ARGS = set()

    @classmethod
    @abstractmethod
    def from_args(cls: type[C], args: A) -> C: ...

    @abstractmethod
    def make_dataset(
        self,
        store: CRDatasetBuilder,
        exs: Sequence[GPExample],
        verbose: bool = False,
    ) -> ColumnarDataset: ...

    @abstractmethod
    def forward(self, *args, **kwargs) -> BaseTorchCanRankOutput: ...

    def rank_dataset(
        self,
        store: CRDatasetBuilder,
        exs: Sequence[GPExample],
        verbose: bool = False,
    ) -> list[TableCanGenUpdateScores]:
        self.eval()

        params = next(self.parameters())
        device = params.device

        dataset = self.make_dataset(store, exs, verbose)
        dloader = DataLoader(
            dataset,
            batch_size=self.EVAL_BATCH_SIZE,
            shuffle=False,
            pin_memory=params.is_cuda,
            collate_fn=dataset.collate_fn,
        )

        with torch.no_grad():
            probs = []
            for batch in tqdm(
                dloader,
                total=len(dloader),
                desc="ranking candidates in a dataset",
                disable=not verbose,
            ):
                kwargs = {}
                for arg in self.EXPECTED_EVAL_ARGS:
                    kwargs[arg] = batch[arg].to(device)
                output = self.forward(**kwargs)
                probs.append(output.probs.cpu())

            probs = torch.cat(probs).numpy()

        example_ranges = dataset.references["example_ranges"]
        return [
            TableCanGenUpdateScores(
                probs[example_ranges[ei][0] : example_ranges[ei][1]]
            )
            for ei, ex in enumerate(exs)
        ]



from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Optional, Protocol, Sequence, TypeVar

import torch
from gp.actors.data import GPExample
from gp.actors.el.cangen import CanGenActor
from gp.entity_linking.candidate_ranking.common import TableCanGenUpdateScores
from gp.entity_linking.candidate_ranking.feats.make_cr_dataset import CRDatasetBuilder
from sm.misc.funcs import import_attr


@dataclass
class PretrainedCanRankModelArgs:
    model_class: str
    model_file: Path | str


class PretrainedCanRankModel:
    VERSION = 100

    def __init__(self, params: PretrainedCanRankModelArgs, cangen_actor: CanGenActor):
        self.params = params
        self.cangen_actor = cangen_actor
        self.dataset_builder = CRDatasetBuilder(cangen_actor)

    def batch_rank_candidates(
        self, exs: Sequence[GPExample], verbose: bool = False
    ) -> list[TableCanGenUpdateScores]:
        return self.method.rank_dataset(self.dataset_builder, exs, verbose)

    @cached_property
    def method(self):
        # load the model
        self.model_class: type[IPretrainedModel] = import_attr(self.params.model_class)
        if torch.cuda.is_available():
            map_location = None
        elif torch.backends.mps.is_available():
            map_location = "mps"
        else:
            map_location = "cpu"
        return self.model_class.load_from_checkpoint(
            self.params.model_file, map_location=map_location
        )


C = TypeVar("C")
A = TypeVar("A")


class IPretrainedModel(Protocol):
    @classmethod
    def load_from_checkpoint(
        cls: type[C], checkpoint: Path | str, map_location: Optional[str] = None
    ) -> C: ...

    def rank_dataset(
        self, store: CRDatasetBuilder, exs: Sequence[GPExample], verbose: bool = False
    ) -> list[TableCanGenUpdateScores]: ...



from __future__ import annotations

import os
import pickle
from abc import abstractmethod
from dataclasses import dataclass
from functools import partial
from typing import Annotated, Dict, Iterable, List, Optional, Sequence, TypeVar

import numpy as np
import orjson
from gp.actors.data import KGDB, KGDBArgs
from gp.misc.appconfig import AppConfig
from hugedict.prelude import RocksDBDict, RocksDBOptions
from libactor.actor import Actor
from loguru import logger
from sm.dataset import Example, FullTable
from sm.misc.matrix import Matrix
from smml.data_model_helper import NumpyDataModel
from tqdm.auto import tqdm

P = TypeVar("P")


@dataclass(slots=True)
class CanEnt:
    id: str
    score: float


class TableCanGenResult(NumpyDataModel):
    __slots__ = ["index", "ent_id", "ent_score"]

    # mapping from col_index => row_index => (start, end)
    index: dict[int, dict[int, tuple[int, int]]]
    ent_id: Annotated[np.ndarray, "String array (Object type)"]
    ent_score: Annotated[np.ndarray, "f64"]

    def __init__(
        self,
        index: dict[int, dict[int, tuple[int, int]]],
        ent_id: Annotated[np.ndarray, "String array (Object type)"],
        ent_score: Annotated[np.ndarray, "f64"],
    ):
        self.index = index
        self.ent_id = ent_id
        self.ent_score = ent_score

    @staticmethod
    def from_matrix(matrix: Matrix[list[CanEnt]]):
        index = {}
        ids = []
        scores = []

        nrows, ncols = matrix.shape()
        for ci in range(ncols):
            index[ci] = {}
            for ri in range(nrows):
                cans = matrix[ri][ci]
                if len(cans) == 0:
                    continue

                index[ci][ri] = (len(ids), len(ids) + len(cans))
                ids.extend([c.id for c in cans])
                scores.extend([c.score for c in cans])

        return TableCanGenResult(
            index=index,
            ent_id=np.asarray(ids, dtype=np.object_),
            ent_score=np.asarray(scores, dtype=np.float64),
        )

    def to_matrix(self, shp: tuple[int, int]) -> Matrix[list[CanEnt]]:
        matrix = Matrix.default(shp, list)
        for ci in self.index:
            for ri in self.index[ci]:
                start, end = self.index[ci][ri]
                matrix[ri][ci] = [
                    CanEnt(self.ent_id[i], self.ent_score[i]) for i in range(start, end)
                ]
        return matrix

    def has_cell_candidates(self, ri: int, ci: int) -> bool:
        return ci in self.index and ri in self.index[ci]

    def get_cell_candidates(self, ri: int, ci: int) -> list[CanEnt]:
        if not self.has_cell_candidates(ri, ci):
            return []
        start, end = self.index[ci][ri]
        return [CanEnt(self.ent_id[i], self.ent_score[i]) for i in range(start, end)]

    def top_k(self, k: int) -> TableCanGenResult:
        matrix = self.to_matrix(self.get_min_shape())
        # using sort instead of heapq.nlargest for stability
        return TableCanGenResult.from_matrix(
            matrix.map(lambda lst: sorted(lst, key=lambda x: x.score, reverse=True)[:k])
        )

    def get_min_shape(self) -> tuple[int, int]:
        """Get the minimum shape of the matrix"""
        ncols = max(self.index.keys()) + 1
        nrows = (
            max(max(row.keys()) if len(row) > 0 else 0 for row in self.index.values())
            + 1
        )
        return nrows, ncols


TableCanGenResult.init()


class CanGenMethod:
    """Candidate ranking method that find candidates from a cell and its surrounding context"""

    @abstractmethod
    def get_candidates(
        self,
        examples: Sequence[Example[FullTable]],
        entity_columns: Sequence[list[int]],
        verbose: bool = False,
    ) -> list[TableCanGenResult]:
        pass


class CanGenBasicMethod(CanGenMethod, Actor[dict]):
    """Candidate generation method that find candidates from text"""

    def __init__(
        self,
        kgdb: KGDB,
        params: dict,
        batch_size: Optional[int] = None,
        soft_limit: int = 1000,
        is_cache_enable: Optional[bool] = None,
    ):
        super().__init__(params, [])
        self.kgdb = kgdb

        self.batch_size = batch_size
        self.soft_limit = soft_limit

        self.method = self.__class__.__name__

        if is_cache_enable is None:
            is_cache_enable = AppConfig.get_instance().is_cache_enable
        if is_cache_enable:
            dbpath = self.actor_dir / "cangen_db"
            dbpath.mkdir(parents=True, exist_ok=True)

            self.kvstore = RocksDBDict(
                path=str(dbpath),
                options=RocksDBOptions(create_if_missing=True),
                deser_key=partial(str, encoding="utf-8"),
                deser_value=pickle.loads,
                ser_value=pickle.dumps,
                # allow to control if the cache should be opened in read-only mode
                # to run multiple experiments in parallel
                readonly=os.environ.get("CANGEN_ROCKSDB_READONLY", "0") == "1",
            )
        else:
            self.kvstore = None

    @abstractmethod
    def get_candidates_by_queries(
        self, queries: List[str]
    ) -> Dict[str, list[tuple[str, float]]]:
        """Generate list of candidate entities for each query."""

    def cached_get_candidates_by_queries(
        self, queries: list[str], cache_only: bool = False
    ) -> dict[str, list[tuple[str, float]]]:
        """Generate list of candidate entities for each query. This should fit to a batch"""
        if self.kvstore is None:
            return self.get_candidates_by_queries(queries)

        output = {
            q: self.kvstore[q][: self.soft_limit] for q in queries if q in self.kvstore
        }
        unk_queries = [q for q in queries if q not in output]
        if not cache_only:
            unk_query_results = self.get_candidates_by_queries(unk_queries)
        else:
            unk_query_results = {q: [] for q in unk_queries}

        for q in unk_queries:
            # sort the results by score and ids so that
            # the order is always deterministic
            res = sorted(unk_query_results[q], key=lambda x: (-x[1], x[0]))
            if not cache_only:
                self.kvstore[q] = res
            output[q] = res[: self.soft_limit]
        return output

    def get_candidates(
        self,
        examples: List[Example[FullTable]],
        entity_columns: list[list[int]],
        verbose: bool = False,
    ) -> list[TableCanGenResult]:
        if verbose:
            logger.debug("Create queries")
        queries = set()
        for ei, example in enumerate(examples):
            for ci in entity_columns[ei]:
                for cell in example.table.table.get_column_by_index(ci).values:
                    queries.add(str(cell))

        if verbose:
            logger.debug("find candidates")
        queries = list(queries)
        if self.kvstore is not None:
            original_queries = queries
            queries = [q for q in queries if q not in self.kvstore]
        else:
            original_queries = []

        if self.batch_size is None:
            batch_size = len(queries)
        else:
            batch_size = self.batch_size

        search_results: dict[str, list[tuple[str, float]]] = {}
        with tqdm(
            total=len(queries), desc="query candidates", disable=not verbose
        ) as pbar:
            for i in range(0, len(queries), batch_size):
                subqueries = queries[i : i + batch_size]
                batch_search_results = self.get_candidates_by_queries(subqueries)
                if self.kvstore is not None:
                    for q in subqueries:
                        # sort the results by score and ids so that
                        # the order is always deterministic
                        res = sorted(
                            batch_search_results[q], key=lambda x: (-x[1], x[0])
                        )
                        self.kvstore[q] = res
                        search_results[q] = res[: self.soft_limit]
                else:
                    for q in subqueries:
                        search_results[q] = batch_search_results[q][: self.soft_limit]

                pbar.update(len(subqueries))

        if self.kvstore is not None:
            with tqdm(
                total=len(original_queries) - len(queries),
                desc="read queries from cache",
                disable=not verbose,
            ) as pbar:
                for q in original_queries:
                    if q not in search_results:
                        search_results[q] = self.kvstore[q][: self.soft_limit]
                        pbar.update(1)

        if verbose:
            logger.debug("find candidates... done! creating the output")

        output = []
        for ei, example in enumerate(examples):
            matrix = Matrix.default(example.table.table.shape(), list)
            for ci in entity_columns[ei]:
                for ri, cell in enumerate(
                    example.table.table.get_column_by_index(ci).values
                ):
                    cell = example.table.table[ri, ci]
                    query = str(cell)
                    query_res = search_results[query]

                    matrix[ri][ci] = [CanEnt(c[0], c[1]) for c in query_res]

            output.append(TableCanGenResult.from_matrix(matrix))

        if verbose:
            logger.debug("creating the output... done!")

        return output


def populate(kgdb: KGDB | KGDBArgs, ids: Iterable[str]):
    db = KGDB.get_instance(kgdb).rudb

    entid: Sequence[str] = []
    label: Sequence[str] = []
    description: Sequence[str] = []
    aliases: Sequence[str] = []
    popularity: Sequence[float] = []

    for id in ids:
        if not db.has_entity(id):
            # entity is not in the database
            newid = db.get_redirected_entity_id(id)
            if newid is None:
                continue
            id = newid
        ent = db.get_entity_metadata(id)

        entid.append(id)
        label.append(ent.label.as_lang_default())
        description.append(ent.description.as_lang_default())
        aliases.append(orjson.dumps(sorted(ent.get_all_aliases())).decode())
        popularity.append(db.get_entity_pagerank(id))

    return entid, label, description, aliases, popularity



from __future__ import annotations

import numpy as np
from gp.actors.data.prelude import GPExample
from gp.entity_linking.candidate_ranking.feats.feat_fn import CRFeatFn, CRFeatFnArgs
from gp.entity_linking.candidate_ranking.feats.get_candidates import GetCandidatesFn


class GetEmptyHeaderEmbeddingFn(CRFeatFn):
    use_args = [
        CRFeatFnArgs.text_embedding_model,
    ]

    get_candidates: GetCandidatesFn

    def __call__(self, ex: GPExample) -> np.ndarray:
        if self.args.text_embedding_model == "sentence-transformers/all-mpnet-base-v2":
            embed_dim = 768
        else:
            raise NotImplementedError(self.args.text_embedding_model)

        n = self.get_candidates(ex).cell.shape[0]
        return np.zeros((n, embed_dim), dtype=np.float32)



"""
Attempt to fix incorrect auto links in the Wikipedia tables:

1. Detect linking to temporal entity and updating it.
"""

from __future__ import annotations

from typing import Literal, Optional

from gp.distantsupervision.make_dataset.mv1 import ChangeStatus, Surgery, TransformV1
from kgdata.models.entity import Entity
from kgdata.models.ont_class import OntologyClass
from kgdata.wikidata.models.wdentity import WDEntity
from sm.dataset import FullTable
from sm.inputs.link import EntityId, Link


class TransformV2(TransformV1):
    """In TransformV1, we map one entity to another entity. However, there are cases where

    we map from one to multiple entities.
    """

    VERSION = 100

    def transform_cell(
        self, table: FullTable, row: int, column: int
    ) -> Optional[list[Link]]:
        """Transform a cell if possible:

        1. empty mentions are ignored even if they are linked (e.g., country flag).
        2. entity that are links to temporal entities are replaced with non-temporal entity.
        """
        non_empty_links = [
            link for link in table.links[row, column] if link.end > link.start
        ]

        modified = len(non_empty_links) != len(table.links[row, column])
        newlinks = []
        for link in non_empty_links:
            newents = []
            for entid in link.entities:
                ent = self.entities[entid]
                newentids, status = self.transform_entities(ent)
                modified = modified or status != ChangeStatus.NoChange
                if status == ChangeStatus.NoChange:
                    newents.append(entid)
                elif status == ChangeStatus.Update:
                    newents.extend(
                        (EntityId(newentid, entid.type) for newentid in newentids)
                    )
                elif status == ChangeStatus.Remove:
                    continue
            if modified:
                newlinks.append(
                    Link(
                        start=link.start,
                        end=link.end,
                        url=link.url,
                        entities=newents,
                    )
                )

        if modified:
            return newlinks
        return None

    def transform_entities(
        self, entity: WDEntity | Entity
    ) -> tuple[list[str], ChangeStatus]:
        """Transform an entity if possible.

        It detects if the entity is an instance of temporal classes such as Football Season, and if found, it will resolve to the correct entity using predefined rules that
        select the value of the property.
        """
        entclasses = [self.classes[cid] for cid in entity.instance_of()]
        if not any(
            any(
                tempcls == entclass.id or tempcls in entclass.ancestors
                for tempcls in self.temporal_classes
            )
            for entclass in entclasses
        ):
            return [entity.id], ChangeStatus.NoChange

        found_surgeries = False
        for surgery in self.known_surgeries:
            newentids, stt = self.try_surgery(surgery, entity, entclasses)
            if stt == "notapplicable":
                continue

            found_surgeries = True

            if stt == "success":
                return newentids, ChangeStatus.Update

            if stt == "nodata":
                # this will print a warning, and we continue to try if there is any other surgery
                # that can be applied. but if there is no surgery that can be applied, we will
                # remove the entity.
                continue

        if not found_surgeries:
            raise Exception(f"Unknown surgery {entity.label} ({entity.id})")
        return [entity.id], ChangeStatus.Remove

    def try_surgery(
        self,
        surgery: Surgery,
        entity: WDEntity | Entity,
        entclasses: list[OntologyClass],
    ) -> tuple[list[str], Literal["nodata", "notapplicable", "success"]]:
        current_ents = [(entity, entclasses)]
        n = len(surgery.steps)

        for i, step in enumerate(surgery.steps):
            matched_current_ents = [
                (current_ent, current_entclasses)
                for current_ent, current_entclasses in current_ents
                if any(
                    step.type == entclass.id or step.type in entclass.ancestors
                    for entclass in current_entclasses
                )
            ]

            if len(matched_current_ents) == 0:
                if i > 0 and not surgery.optional:
                    self.logger.warning(
                        "The next step requires type {}, but entities {} isn't of that type.",
                        step.type,
                        ", ".join(
                            f"`{current_ent.label} ({current_ent.id})`"
                            for current_ent, _ in current_ents
                        ),
                    )
                    return [], "nodata"
                return [], "notapplicable"

            matched_current_ents = [
                (current_ent, current_entclasses)
                for current_ent, current_entclasses in matched_current_ents
                if step.prop in current_ent.props
            ]

            if len(matched_current_ents) == 0:
                if not surgery.optional:
                    self.logger.warning(
                        "Missing {} for {}",
                        step.prop,
                        ", ".join(
                            f"`{current_ent.label} ({current_ent.id})`"
                            for current_ent, _ in current_ents
                        ),
                    )
                return [], "nodata"

            potential_values = sorted(
                {
                    val
                    for current_ent, _ in matched_current_ents
                    for val in current_ent.get_object_prop_value(step.prop)
                }
            )
            if i == n - 1:
                return potential_values, "success"

            current_ents = [
                (
                    (ent := self.entities[eid]),
                    [self.classes[cid] for cid in ent.instance_of()],
                )
                for eid in potential_values
            ]

        raise Exception("Shouldn't reach here")



from __future__ import annotations

import re
from dataclasses import dataclass, field
from typing import Optional, Sequence

from gp.distantsupervision.make_dataset.utils import FilterMixin, StrPath
from sm.dataset import FullTable

# fmt: off
DEFAULT_FILTER_PATTERNS = [
    r"\d+",      # number year 1920, etc.
    r"\d+-\d+",  # number range 1920-1925 or year-month 1920-01, etc.
    r"\[\d+\]",  # references [1], [2], etc.
]
# fmt: on


@dataclass
class FilterRegexArgs:
    noisy_threshold: float = 0.5
    patterns: Sequence[str] = field(default_factory=lambda: DEFAULT_FILTER_PATTERNS)


class FilterRegex(FilterMixin):
    VERSION = 101

    def __init__(self, args: FilterRegexArgs, log_file: Optional[StrPath] = None):
        super().__init__(log_file)
        self.patterns = args.patterns
        self.regexes = [re.compile(p) for p in self.patterns]

    def is_noisy(self, table: FullTable, ci: int) -> tuple[bool, str]:
        for cell in table.table.get_column_by_index(ci).values:
            for regex in self.regexes:
                if regex.match(cell):
                    return True, f"cell {cell} matches {regex.pattern}"
        return False, ""



import re
from dataclasses import dataclass
from pathlib import Path
from typing import Mapping, Optional

import serde.json
import serde.yaml
from gp.distantsupervision.make_dataset.interface import LabelFn
from gp.distantsupervision.make_dataset.lv1 import LabelV1
from gp.distantsupervision.make_dataset.utils import FilterMixin, StrPath
from kgdata.models.ont_class import OntologyClass
from sm.dataset import FullTable

DEFAULT_WHITELIST_FILENAME = "whitelist.yml"


@dataclass
class FilterByHeaderColTypeArgs:
    whitelist_file: Path


class FilterByHeaderColType(FilterMixin):
    """This function is deprecated. Use a labeler that combines the whitelist instead"""

    VERSION = 101

    def __init__(
        self,
        args: FilterByHeaderColTypeArgs,
        labeler: LabelFn,
        log_file: Optional[StrPath] = None,
    ):
        super().__init__(log_file)
        # mapping from class id to list of allowed column names
        self.whitelist: dict[str, set[str]] = self.parse_whitelist(args.whitelist_file)
        self.whitelist_keys = set(self.whitelist.keys())
        assert isinstance(labeler, LabelV1), f"Expected LabelV1, got {type(labeler)}"
        self.labeler = labeler
        self.classes = labeler.classes

    def is_noisy(self, table: FullTable, ci: int) -> tuple[bool, str]:
        header = table.table.get_column_by_index(ci).clean_multiline_name
        if header is None:
            return True, "no header"
        col_types = self.labeler.label_column(table.links[:, ci])
        for col_type in col_types:
            if col_type.id in self.whitelist_keys:
                if header in self.whitelist[col_type.id]:
                    return False, ""
            else:
                cls = self.classes[col_type.id]
                common_ancestors = self.whitelist_keys.intersection(
                    cls.ancestors.keys()
                )
                if any(
                    header in self.whitelist[ancestor] for ancestor in common_ancestors
                ):
                    return False, ""
        return (
            True,
            f"Column {header} with candidate types {[f'{self.classes[ctype.id].label} ({ctype.id})' for ctype in col_types]} not in whitelist",
        )

    def parse_whitelist(self, infile: Path):
        whitelist: dict[str, set[str]] = {}
        label_parser = re.compile(r"[^(]+ \((Q\d+)\)")
        if infile.suffix in {".yml", ".yaml"}:
            label2headers = serde.yaml.deser(infile)
        elif infile.suffix in {".json"}:
            label2headers = serde.json.deser(infile)
        else:
            raise ValueError(f"Unknown whitelist file format: {infile.suffix}")

        for label, headers in label2headers.items():
            m = label_parser.match(label)
            assert m is not None
            class_id = m.group(1)
            whitelist[class_id] = set(headers)

        return whitelist



from __future__ import annotations

from typing import Any, Sequence

import ray
from gp.actors.data import KGDB, KGDBArgs
from kgdata.wikidata.datasets import entity_pagerank
from ream.cache_helper import Cache, MemBackend
from sm.misc.ray_helper import (
    add_ray_actors,
    get_ray_actors,
    is_parallelizable,
    ray_actor_map,
)


class KGDBService:
    def __init__(self, kgdb: KGDB | KGDBArgs):
        if not isinstance(kgdb, KGDB):
            kgdb = KGDB(kgdb)
        self.kgdb = kgdb

    @Cache.flat_cache(backend=MemBackend(), cache_key=lambda self, key, ids: key)
    def labels(self, key: str, ids: Sequence[str]) -> list[str]:
        db = self.kgdb.pydb.entity_labels.cache()
        return [db[id] for id in ids]

    @Cache.flat_cache(backend=MemBackend(), cache_key=lambda self, key, ids: key)
    def aliases(self, key: str, ids: Sequence[str]) -> list[list[str]]:
        db = self.kgdb.pydb.entity_metadata.cache()
        return [sorted(db[id].aliases.flatten()) for id in ids]

    @Cache.flat_cache(backend=MemBackend(), cache_key=lambda self, key, ids: key)
    def pageranks(self, key: str, ids: Sequence[str]) -> list[float]:
        db = self.kgdb.pydb.entity_pagerank.cache()
        return [db[id] for id in ids]


class KGDBBatchService:
    def __init__(self, kgdb: KGDBArgs):
        self.kgdb = kgdb
        self.ns = "gp.entity_linking.populate.KGDBService:" + kgdb.get_key()

    @Cache.flat_cache(backend=MemBackend(), cache_key=lambda self, key_ids: key_ids[0])
    def batch_labels(
        self, lst_key_ids: list[tuple[str, Sequence[str]]]
    ) -> list[list[str]]:
        return ray_actor_map(
            [actor.labels.remote for actor in self.get_actors()],
            lst_key_ids,
        )

    @Cache.flat_cache(backend=MemBackend(), cache_key=lambda self, key_ids: key_ids[0])
    def batch_aliases(
        self, lst_key_ids: list[tuple[str, Sequence[str]]]
    ) -> list[list[list[str]]]:
        return ray_actor_map(
            [actor.aliases.remote for actor in self.get_actors()],
            lst_key_ids,
        )

    @Cache.flat_cache(backend=MemBackend(), cache_key=lambda self, key_ids: key_ids[0])
    def batch_pageranks(
        self, lst_key_ids: list[tuple[str, Sequence[str]]]
    ) -> list[list[float]]:
        return ray_actor_map(
            [actor.pageranks.remote for actor in self.get_actors()],
            lst_key_ids,
        )

    def get_actors(self) -> list["ray.ObjectRef[KGDBService]"]:
        actors = get_ray_actors(self.ns)
        if len(actors) == 0:
            actors = add_ray_actors(
                KGDBService, (self.kgdb,), self.ns, size=6, scope=-1
            )
        return actors



from __future__ import annotations

from dataclasses import dataclass, field
from functools import cached_property
from typing import Optional

from gp.actors.data import KGDB
from gp.entity_linking.candidate_generation.common import TableCanGenResult
from gp.misc.conversions import to_rust_table
from gp.semanticmodeling.cangraph_cfg import CanGraphExtractorCfg
from gp.semanticmodeling.text_parser import TextParser, TextParserConfigs
from gp_core.algorithms import (
    CanGraphExtractedResult,
    extract_cangraph,
    par_extract_cangraphs,
)
from gp_core.models import TableCells
from libactor.actor import Actor
from libactor.cache import (
    BackendFactory,
    BorrowBackend,
    FlatCacheArg,
    IdentObj,
    cache,
    flat_cache,
)
from loguru import logger
from sm.dataset import Example, FullTable
from timer import Timer
from tqdm import tqdm


@dataclass
class CanGraphActorArgs:
    topk: Optional[int] = field(
        default=100,
        metadata={
            "help": "keeping maximum top k candidates for each cell",
        },
    )
    text_parser: TextParserConfigs = field(
        default_factory=TextParserConfigs,
        metadata={"help": "Configuration for the text parser"},
    )
    cangraph_extractor: CanGraphExtractorCfg = field(
        default_factory=CanGraphExtractorCfg,
        metadata={"help": "Configuration for the candidate graph extractor"},
    )


class CanGraphActor(Actor[CanGraphActorArgs]):
    VERSION = 117

    def __init__(self, params: CanGraphActorArgs):
        super().__init__(params, dep_actors=[])
        self.batch_invoke_cache = {}

    @cached_property
    def text_parser(self) -> TextParser:
        return TextParser(self.params.text_parser)

    @cached_property
    def ru_cangraph_extractor(self):
        return self.params.cangraph_extractor.to_rust()

    @cache(
        backend=BackendFactory.actor.sqlite.pickle(
            filename="extract_cangraph.sqlite", mem_persist=True
        ),
        cache_args=["example", "can_ent", "kgdb"],
    )
    def invoke(
        self,
        example: IdentObj[Example[FullTable]],
        can_ent: IdentObj[TableCanGenResult],
        kgdb: IdentObj[KGDB],
        parallel: bool = True,
    ) -> CanGraphExtractedResult:
        if (example.key, can_ent.key) in self.batch_invoke_cache:
            return self.batch_invoke_cache[example.key, can_ent.key]

        ex = example.value
        nrows, ncols = ex.table.table.shape()
        text_parser = self.text_parser

        return extract_cangraph(
            to_rust_table(ex, can_ent.value),
            TableCells(
                [
                    [
                        text_parser.parse(ex.table.table[ri, ci]).to_rust()
                        for ci in range(ncols)
                    ]
                    for ri in range(nrows)
                ]
            ),
            kgdb.value.rudb,
            self.ru_cangraph_extractor,
            None,
            parallel=parallel,
        )

    # @flat_cache(backend=BorrowBackend(invoke), cache_args=[])
    def batch_invoke(
        self,
        exs: list[IdentObj[Example[FullTable]]],
        can_ents: list[IdentObj[TableCanGenResult]],
        kgdb: IdentObj[KGDB],
        verbose: bool = True,
    ):
        if len(exs) == 0:
            return []

        text_parser = self.text_parser

        tables = []
        table_cells = []

        for ei, ex in tqdm(enumerate(exs), disable=not verbose):
            nrows, ncols = ex.value.table.table.shape()
            tbl = to_rust_table(ex.value, can_ents[ei].value)
            cells = TableCells(
                [
                    [
                        text_parser.parse(ex.value.table.table[ri, ci]).to_rust()
                        for ci in range(ncols)
                    ]
                    for ri in range(nrows)
                ]
            )

            tables.append(tbl)
            table_cells.append(cells)

        with Timer().watch_and_report(
            "Extracting candidate graphs", print_fn=logger.info, preprint=True
        ):
            cangraphs = par_extract_cangraphs(
                tables,
                table_cells,
                kgdb.value.rudb,
                self.ru_cangraph_extractor,
                None,
                verbose=verbose,
            )

        with Timer().watch_and_report(
            "Caching candidate graphs", print_fn=logger.info, preprint=True
        ):
            self.batch_invoke_cache = {}
            for ex, can_ent, cangraph in zip(exs, can_ents, cangraphs):
                self.batch_invoke_cache[ex.key, can_ent.key] = cangraph
                self.invoke(ex, can_ent, kgdb, False)
            self.batch_invoke_cache = {}
        return cangraphs

    # @Cache.flat_cache(
    #     backend=Cache.sqlite.pickle(
    #         filename="cangraph", mem_persist=True, compression="lz4"
    #     ),
    #     cache_key=lambda self, example, verbose=False: example.id,
    #     disable=lambda self: not AppConfig.get_instance().is_cache_enable,
    # )
    # def batch_call(
    #     self, exs: list[GPExample], verbose: bool = False
    # ) -> list[CanGraphExtractedResult]:
    #     if len(exs) == 0:
    #         return []

    #     assert all(
    #         ex.kgname == exs[0].kgname for ex in exs
    #     ), "Don't support multiple KGs in the same function call"

    #     ex_cans = self.get_candidate_entities(exs)
    #     text_parser = self.text_parser

    #     tables = []
    #     table_cells = []

    #     for ei, ex in tqdm(enumerate(exs), disable=not verbose):
    #         nrows, ncols = ex.table.table.shape()
    #         tbl = to_rust_table(ex, ex_cans[ei])
    #         cells = TableCells(
    #             [
    #                 [
    #                     text_parser.parse(ex.table.table[ri, ci]).to_rust()
    #                     for ci in range(ncols)
    #                 ]
    #                 for ri in range(nrows)
    #             ]
    #         )

    #         tables.append(tbl)
    #         table_cells.append(cells)

    #     return par_extract_cangraphs(
    #         tables,
    #         table_cells,
    #         self.db_actor.kgdbs[exs[0].kgname].rudb,
    #         self.ru_cangraph_extractor,
    #         None,
    #         verbose=verbose,
    #     )

    # def get_candidate_entities(self, exs: list[GPExample]):
    #     ex_cans = self.canrank_actor.batch_call(exs)
    #     if self.params.topk is not None:
    #         return [cans.top_k(self.params.topk) for cans in ex_cans]
    #     return ex_cans



from __future__ import annotations

from abc import abstractmethod
from pathlib import Path
from typing import Optional, Union

from gp.distantsupervision.make_dataset.interface import FilterFn
from loguru import logger
from ream.helper import _logger_formatter
from sm.dataset import FullTable

StrPath = Union[str, Path]


class LoggerMixin:
    def __init__(self, log_file: Optional[StrPath] = None):
        self.logger = logger.bind(name=self.__class__.__name__)
        if log_file is not None:
            self.logger.add(log_file, format=_logger_formatter)


class FilterMixin(FilterFn, LoggerMixin):
    def filter(self, table: FullTable, columns: list[int]) -> list[int]:
        new_cols = []
        for ci in columns:
            noisy, reason = self.is_noisy(table, ci)
            if not noisy:
                new_cols.append(ci)
            else:
                self.logger.debug(
                    f"Filter column {ci} in table {table.table.table_id}: {reason}"
                )
        return new_cols

    @abstractmethod
    def is_noisy(self, table: FullTable, ci: int) -> tuple[bool, str]:
        pass



from .erv1 import EntityRecognitionV1, EntityRecognitionV1Args
from .fbhct import FilterByHeaderColType, FilterByHeaderColTypeArgs
from .fbt import FilterByEntType, FilterByEntTypeArgs
from .fcom import CombinedFilter, CombinedFilterArgs
from .fentcol import FilterNotEntCol
from .freg import FilterRegex, FilterRegexArgs
from .fv1 import FilterV1, FilterV1Args
from .interface import FilterFn, LabelFn, NoFilter, NoTransform, TransformFn
from .lv1 import LabelV1, LabelV1Args
from .lv2 import LabelV2, LabelV2Args
from .mv1 import TransformV1, TransformV1Args
from .mv2 import TransformV2
from .utils import FilterMixin

__all__ = [
    "FilterFn",
    "LabelFn",
    "NoFilter",
    "FilterMixin",
    "FilterRegex",
    "FilterRegexArgs",
    "FilterV1",
    "FilterV1Args",
    "FilterByEntType",
    "FilterByEntTypeArgs",
    "FilterByHeaderColType",
    "FilterByHeaderColTypeArgs",
    "FilterNotEntCol",
    "CombinedFilter",
    "CombinedFilterArgs",
    "TransformFn",
    "NoTransform",
    "TransformV1",
    "TransformV1Args",
    "TransformV2",
    "LabelV1",
    "LabelV1Args",
    "LabelV2",
    "LabelV2Args",
    "EntityRecognitionV1",
    "EntityRecognitionV1Args",
]



from dataclasses import dataclass

from sm.dataset import FullTable


@dataclass
class EntityRecognitionV1Args:
    min_link_freq: float = 0.7
    min_link_coverage: float = 0.9


class EntityRecognitionV1:
    VERSION = 100

    def __init__(self, args: EntityRecognitionV1Args):
        self.args = args

    def recognize(self, table: FullTable) -> list[int]:
        return [
            col.index
            for col in table.table.columns
            if self.is_entity_column(table, col.index)
        ]

    def is_entity_column(self, table: FullTable, ci: int) -> bool:
        cells = table.table.get_column_by_index(ci).values

        n_rows = len(cells)
        n_links = 0
        coverage_size = 0
        total_area = 0

        for ri, cell in enumerate(cells):
            links = table.links[ri, ci]
            if len(links) == 0:
                continue

            coverage_size += sum((link.end - link.start) for link in links)
            total_area += len(cell)

            n_links += 1

        return (
            n_links / n_rows >= self.args.min_link_freq
            and coverage_size / total_area >= self.args.min_link_coverage
        )



from __future__ import annotations

import re
from dataclasses import dataclass, field
from typing import Optional, Sequence

from gp.distantsupervision.make_dataset.fbhct import FilterByHeaderColTypeArgs
from gp.distantsupervision.make_dataset.fbt import FilterByEntTypeArgs
from gp.distantsupervision.make_dataset.freg import FilterRegexArgs
from gp.distantsupervision.make_dataset.utils import FilterMixin, StrPath
from sm.dataset import FullTable


@dataclass
class CombinedFilterArgs:
    regex: FilterRegexArgs = field(
        default_factory=FilterRegexArgs,
    )
    ignore_types: FilterByEntTypeArgs = field(
        default_factory=FilterByEntTypeArgs,
    )
    header_col_type: Optional[FilterByHeaderColTypeArgs] = field(default=None)


class CombinedFilter(FilterMixin):
    VERSION = 100

    def __init__(
        self, filters: Sequence[FilterMixin], log_file: Optional[StrPath] = None
    ):
        super().__init__(log_file)
        self.filters = filters

    def is_noisy(self, table: FullTable, ci: int) -> tuple[bool, str]:
        for f in self.filters:
            is_noisy, reason = f.is_noisy(table, ci)
            if is_noisy:
                return True, reason
        return False, ""



from __future__ import annotations

import re
from abc import abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Literal, Optional, get_type_hints

from gp.actors.data.prelude import GPExample
from gp.misc.appconfig import AppConfig
from gramsplus.semanticmodeling.literal_matcher import LiteralMatcherConfig
from gramsplus.semanticmodeling.text_parser import TextParserConfigs
from ream.cache_helper import CacheableFn, MemBackend, assign_dataclass_field_names

if TYPE_CHECKING:
    from gp.entity_linking.candidate_ranking.feats.make_cr_dataset import CRDatasetBuilder


@dataclass
class CRFeatFnArgs:
    topk: Optional[int] = field(
        default=None,
        metadata={"help": "The number of candidates to keep for each cell."},
    )
    topk_scoring: Literal["default"] = field(
        default="default",
        metadata={"help": "The scoring method for topk candidates."},
    )
    add_missing_gold: bool = field(
        default=False,
        metadata={
            "help": "Whether to add missing gold entities if candidate retrieval misses them"
        },
    )
    remove_nil_entity: bool = field(
        default=False,
        metadata={"help": "Whether to remove NIL entity from candidates"},
    )
    text_parser_cfg: TextParserConfigs = field(
        default_factory=TextParserConfigs,
        metadata={"help": "The text parser configuration"},
    )
    literal_matcher_cfg: LiteralMatcherConfig = field(
        default_factory=LiteralMatcherConfig,
        metadata={"help": "The literal matcher configuration"},
    )
    text_embedding_model: str = field(
        default="",
        metadata={"help": "The embedding model to calculate text embeddings"},
    )


assign_dataclass_field_names(CRFeatFnArgs)


class CRFeatFn(CacheableFn):
    backends = []
    use_args = []

    def __init__(self, store: CRDatasetBuilder):
        super().__init__(
            self.use_args,
            store.get_working_fs(),
            not AppConfig.get_instance().is_cache_enable,
        )
        # to assign functions based on type hints
        for name, type in get_type_hints(self.__class__).items():
            if issubclass(type, CacheableFn):
                setattr(self, name, store.get_func(type))
        self.store = store
        self.cangen_actor = store.cangen_actor
        self.args: CRFeatFnArgs = CRFeatFnArgs()

    @abstractmethod
    def __call__(self, ex: GPExample) -> Any:
        """This is where to put the function body. To cache it, wraps it with @Cache.<X> decorators"""
        ...

    def batch_call(self, exs: list[GPExample]) -> list[Any]:
        return [self(ex) for ex in exs]

    def set_args(self, args: CRFeatFnArgs):
        self.args = args
        self.cache_file = self.get_cache_file()
        return self

    def get_cache_file(self):
        if hasattr(self, "VERSION"):
            version = f"_v{getattr(self, 'VERSION')}"
        else:
            version = ""

        filename = f"%s%s" % (
            re.sub(
                r"(?<!^)(?=[A-Z])", "_", self.__class__.__name__.replace("Fn", "")
            ).lower(),
            version,
        )

        fs = self.get_working_fs()
        file = fs.get(filename, key=self.get_cache_key(self, self.args), save_key=True)
        if file.exists():
            return str(file.get().relative_to(fs.root))

        with fs.acquire_write_lock(), file.reserve_and_track() as realfile:
            return str(realfile.relative_to(fs.root))

    @classmethod
    def new_mem_backend(cls):
        backend = MemBackend()
        cls.backends.append(backend)
        return backend

    @classmethod
    @contextmanager
    def auto_clear_mem_backend(cls):
        try:
            yield None
        finally:
            for backend in cls.backends:
                backend.clear()


# class GetTopKCanFn(CRFeatFn):
#     use_args = [CRFeatFnArgs.topk, CRFeatFnArgs.topk_scoring]
#     get_candidates: GetCandidatesFn
#     get_canbase: GetCanBaseFn
#     get_can_basefeat: GetCanBaseFeatFn

#     @Cache.cache(
#         backend=Cache.cls.dir(
#             cls=[DatasetCandidateEntities, SingleNumpyArray],
#             mem_persist=CRFeatFn.new_mem_backend(),
#             dirname=CRFeatFn.get_dirname,
#         ),
#         **CRFeatFn.cache_decorator_args(),
#     )
#     def __call__(
#         self, args: CRFeatFnArgs
#     ) -> tuple[DatasetCandidateEntities, Optional[SingleNumpyArray]]:
#         cans = self.get_candidates(args)
#         topk_index = None

#         if args.topk is not None:
#             assert args.topk_scoring == "default"

#             UPPERBOUND_SCORE = 1e6
#             cans.score[cans.score == GetCandidatesFn.DEFAULT_GOLD_IMPUTE_SCORE] = (
#                 UPPERBOUND_SCORE
#             )
#             cans, topk_index = cans.top_k_candidates(args.topk, return_index_remap=True)
#             cans.score[cans.score == UPPERBOUND_SCORE] = (
#                 GetCandidatesFn.DEFAULT_GOLD_IMPUTE_SCORE
#             )
#             topk_index = SingleNumpyArray(topk_index)

#         return cans, topk_index


# class GetTopKCanBaseFn(CRFeatFn):
#     get_canbase: GetCanBaseFn
#     get_topk_candidates: GetTopKCanFn

#     @Cache.cache(
#         backend=CRFeatFn.new_mem_backend(),
#         cache_key=CacheableFn.get_cache_key,
#     )
#     def __call__(self, args: CRFeatFnArgs) -> CRDatasetCan:
#         subset_index = self.get_topk_candidates(args)[1]
#         canbase = self.get_canbase(args)
#         if subset_index is not None:
#             canbase = canbase.select(subset_index.value)
#         return canbase


# class GetTopKCanBaseFeatFn(CRFeatFn):
#     get_can_basefeat: GetCanBaseFeatFn
#     get_topk_candidates: GetTopKCanFn

#     @Cache.cache(
#         backend=CRFeatFn.new_mem_backend(),
#         cache_key=CacheableFn.get_cache_key,
#     )
#     def __call__(self, args: CRFeatFnArgs) -> CRDatasetFeatures:
#         subset_index = self.get_topk_candidates(args)[1]
#         can_basefeat = self.get_can_basefeat(args)
#         if subset_index is not None:
#             can_basefeat = can_basefeat.select(subset_index.value)
#         return can_basefeat


# class GetHeaderEmbeddingFn(CRFeatFn):
#     use_args = [
#         CRFeatFnArgs.dsquery,
#         CRFeatFnArgs.text_embedding_model,
#     ]
#     get_topk_candidates: GetTopKCanFn

#     @Cache.cache(
#         backend=CRFeatFn.new_mem_backend(),
#         cache_key=CacheableFn.get_cache_key,
#     )
#     def __call__(self, args: CRFeatFnArgs) -> np.ndarray:
#         col_names = self.get_headers(args)
#         return self.store.get_text_embedding(args.text_embedding_model).batch_get(
#             col_names, verbose=AppConfig.get_instance().is_canrank_verbose
#         )

#     @Cache.cache(
#         backend=CRFeatFn.new_mem_backend(),
#         cache_key=CacheableFn.get_cache_key,
#     )
#     def get_headers(self, args: CRFeatFnArgs) -> BatchText:
#         examples = self.data_actor(args.dsquery)
#         topkcans = self.get_topk_candidates(args)[0]

#         col_names = BatchText()
#         for ex in examples:
#             for col_index, (cstart, cend, _) in topkcans.index[ex.table.table.table_id][
#                 2
#             ].items():
#                 col = ex.table.table.get_column_by_index(col_index)
#                 col_name = assert_not_null(col.clean_multiline_name)
#                 col_names.add_repeated_text(col_name, cend - cstart)

#         assert len(topkcans) == len(col_names)
#         return col_names


# class GetEmptyHeaderEmbeddingFn(CRFeatFn):
#     use_args = [
#         CRFeatFnArgs.dsquery,
#         CRFeatFnArgs.text_embedding_model,
#     ]
#     get_topk_candidates: GetTopKCanFn

#     @Cache.cache(
#         backend=CRFeatFn.new_mem_backend(),
#         cache_key=CacheableFn.get_cache_key,
#     )
#     def __call__(self, args: CRFeatFnArgs) -> np.ndarray:
#         if args.text_embedding_model == "sentence-transformers/all-mpnet-base-v2":
#             embed_dim = 768
#         else:
#             raise NotImplementedError(args.text_embedding_model)

#         topkcans = self.get_topk_candidates(args)[0]
#         return np.zeros((len(topkcans), embed_dim), dtype=np.float32)


# class GetAutoHeaderEmbeddingFn(CRFeatFn):
#     use_args = [
#         CRFeatFnArgs.dsquery,
#         CRFeatFnArgs.text_embedding_model,
#     ]
#     get_topk_candidates: GetTopKCanFn
#     get_header_embedding: GetHeaderEmbeddingFn

#     @Cache.cache(
#         backend=CRFeatFn.new_mem_backend(),
#         cache_key=CacheableFn.get_cache_key,
#     )
#     def __call__(self, args: CRFeatFnArgs) -> np.ndarray:
#         if args.text_embedding_model == "sentence-transformers/all-mpnet-base-v2":
#             embed_dim = 768
#         else:
#             raise NotImplementedError(args.text_embedding_model)

#         col_names = self.get_header_embedding.get_headers(args)
#         embeds = self.get_header_embedding(args)

#         empty_header = []
#         for name, idx in col_names.unique_text.items():
#             # for each column, if the header is "" or "colX", we convert it to empty embed
#             if name.strip() == "" or re.match(r"col\d+", name) is not None:
#                 empty_header.append(idx)

#         embeds[empty_header] = np.zeros(
#             (len(empty_header), embed_dim), dtype=np.float32
#         )
#         return embeds


# class HasHeaderFn(CRFeatFn):
#     use_args = [
#         CRFeatFnArgs.dsquery,
#         CRFeatFnArgs.text_embedding_model,
#     ]
#     get_topk_candidates: GetTopKCanFn
#     get_header_embedding: GetHeaderEmbeddingFn

#     @Cache.cache(
#         backend=Cache.cls.dir(
#             cls=SingleNumpyArray,
#             mem_persist=CRFeatFn.new_mem_backend(),
#             dirname=CRFeatFn.get_dirname,
#         ),
#         **CRFeatFn.cache_decorator_args(),
#     )
#     def __call__(self, args: CRFeatFnArgs) -> SingleNumpyArray:
#         col_names = self.get_header_embedding.get_headers(args)
#         empty_header = set()
#         for name, idx in col_names.unique_text.items():
#             # for each column, if the header is "" or "colX", we convert it to empty embed
#             if name.strip() == "" or re.match(r"col\d+", name) is not None:
#                 empty_header.add(idx)

#         return SingleNumpyArray(
#             np.asarray(
#                 [i not in empty_header for i in col_names.text_index], dtype=np.bool_
#             )
#         )


# class HasHeader2Fn(CRFeatFn):
#     use_args = [
#         CRFeatFnArgs.dsquery,
#         CRFeatFnArgs.text_embedding_model,
#     ]
#     get_topk_candidates: GetTopKCanFn
#     get_header_embedding: GetHeaderEmbeddingFn

#     @Cache.cache(
#         backend=Cache.cls.dir(
#             cls=SingleNumpyArray,
#             mem_persist=CRFeatFn.new_mem_backend(),
#             dirname=CRFeatFn.get_dirname,
#         ),
#         **CRFeatFn.cache_decorator_args(),
#     )
#     def __call__(self, args: CRFeatFnArgs) -> SingleNumpyArray:
#         col_names = self.get_header_embedding.get_headers(args)
#         empty_header = set()
#         for name, idx in col_names.unique_text.items():
#             # for each column, if the header is "" or "colX", we convert it to empty embed
#             if re.match(r"col\d+", name) is not None:
#                 empty_header.add(idx)

#         return SingleNumpyArray(
#             np.asarray(
#                 [i not in empty_header for i in col_names.text_index], dtype=np.bool_
#             )
#         )


# class GetEntityDescriptionEmbeddingFn(CRFeatFn):
#     use_args = [
#         CRFeatFnArgs.dsquery,
#         CRFeatFnArgs.text_embedding_model,
#     ]
#     get_topk_candidates: GetTopKCanFn

#     @Cache.cache(
#         backend=CRFeatFn.new_mem_backend(),
#         cache_key=CacheableFn.get_cache_key,
#     )
#     def __call__(self, args: CRFeatFnArgs) -> np.ndarray:
#         topkcans = self.get_topk_candidates(args)[0]

#         texts = BatchText.from_list_str(topkcans.description)
#         return self.store.get_text_embedding(args.text_embedding_model).batch_get(
#             texts, verbose=AppConfig.get_instance().is_canrank_verbose
#         )


# class GetEmptyEntityDescriptionEmbeddingFn(CRFeatFn):
#     use_args = [
#         CRFeatFnArgs.dsquery,
#         CRFeatFnArgs.text_embedding_model,
#     ]
#     get_topk_candidates: GetTopKCanFn

#     @Cache.cache(
#         backend=CRFeatFn.new_mem_backend(),
#         cache_key=CacheableFn.get_cache_key,
#     )
#     def __call__(self, args: CRFeatFnArgs) -> np.ndarray:
#         if args.text_embedding_model == "sentence-transformers/all-mpnet-base-v2":
#             embed_dim = 768
#         else:
#             raise NotImplementedError(args.text_embedding_model)

#         topkcans = self.get_topk_candidates(args)[0]
#         return np.zeros((len(topkcans), embed_dim), dtype=np.float32)


# if __name__ == "__main__":
#     import IPython

#     IPython.embed()
#     print(CRFeatFnArgs.dsquery)



from __future__ import annotations

from abc import ABC, abstractmethod

from sm.dataset import FullTable
from sm.inputs.link import EntityIdWithScore


class FilterFn(ABC):
    @abstractmethod
    def filter(self, table: FullTable, columns: list[int]) -> list[int]:
        ...


class TransformFn(ABC):
    @abstractmethod
    def transform(self, table: FullTable, columns: list[int]) -> FullTable:
        ...


class LabelFn(ABC):
    @abstractmethod
    def label(
        self, table: FullTable, columns: list[int]
    ) -> list[list[EntityIdWithScore]]:
        ...


class NoFilter(FilterFn):
    def filter(self, table: FullTable, columns: list[int]) -> list[int]:
        return columns


class NoTransform(TransformFn):
    def transform(self, table: FullTable, columns: list[int]) -> FullTable:
        return table



from __future__ import annotations

from abc import abstractmethod
from copy import copy
from dataclasses import dataclass
from functools import cached_property
from operator import itemgetter
from pathlib import Path
from typing import Optional

import lightning.pytorch as pl
import torch
from gp.misc.dataset import ColumnarDataset
from gp.semanticmodeling.postprocessing.interface import EdgeProb, NodeProb
from ream.actors.base import BaseActor
from ream.cache_helper import Cache, MemBackend
from ream.data_model_helper import NumpyDataModelHelper
from ream.dataset_helper import DatasetList
from ream.helper import import_attr

try:
    from gp.actors.candidate_graph import CanGraphActor
    from scripts.www24.cpa_helper import make_dataset
    from scripts.www24.cta_helper import predict_column_prob, predict_cta
    from scripts.www24.stores.cpa_store import CPAStore
    from scripts.www24.stores.cpa_store_fn import CPAStoreFn, CPAStoreFnArgs
except ImportError:
    print("TODO: fix imports on this file")

from sm.misc.funcs import import_attr
from torch.utils.data import DataLoader
from tqdm.auto import tqdm


@dataclass
class CPAStoreInvokeArgs:
    args: CPAStoreFnArgs
    funcs: list[str]

    @cached_property
    def enable_fns(self) -> list[type[CPAStoreFn]]:
        return [import_attr(fn) for fn in self.funcs]


@dataclass
class PretrainedCPAModelArgs:
    model_class: str
    model_file: Path
    store_invoke_args: CPAStoreInvokeArgs


class PretrainedCPAModel(BaseActor[PretrainedCPAModelArgs]):
    VERSION = 100

    def __init__(
        self,
        params: PretrainedCPAModelArgs,
        store: CPAStore,
    ):
        super().__init__(params, [store])
        self.store = store

    @Cache.cache(backend=MemBackend())
    def __call__(self, dsquery: str, verbose: bool) -> DatasetList[EdgeProb]:
        args = copy(self.params.store_invoke_args.args)
        args.dsquery = dsquery
        return self.method.predict_dataset(
            self.store,
            CPAStoreInvokeArgs(args, self.params.store_invoke_args.funcs),
            verbose,
        )

    @cached_property
    def method(self):
        self.model_class: type[BaseTorchCPAModel] = import_attr(self.params.model_class)
        if torch.cuda.is_available():
            map_location = "cuda"
        else:
            map_location = "cpu"
        return self.model_class.load_from_checkpoint(
            self.params.model_file, map_location=map_location
        )


class BaseTorchCPAModel(pl.LightningModule):
    EVAL_BATCH_SIZE = 100
    EXPECTED_EVAL_ARGS = set()

    def make_dataset(
        self, store: CPAStore, store_invoke_args: CPAStoreInvokeArgs
    ) -> ColumnarDataset:
        ds = store(store_invoke_args.args, store_invoke_args.enable_fns)
        return make_dataset(ds, is_train=False)

    def predict_dataset(
        self, store: CPAStore, store_invoke_args: CPAStoreInvokeArgs, verbose: bool
    ) -> DatasetList[EdgeProb]:
        self.eval()

        params = next(self.parameters())
        device = params.device

        dsquery = store_invoke_args.args.dsquery

        dataset = self.make_dataset(store, store_invoke_args)
        dloader = DataLoader(
            dataset,
            batch_size=self.EVAL_BATCH_SIZE,
            shuffle=False,
            pin_memory=params.is_cuda,
            collate_fn=dataset.collate_fn,
        )
        with torch.no_grad():
            probs = []
            for batch in tqdm(
                dloader,
                total=len(dloader),
                desc="predicting cpa",
                disable=not verbose,
            ):
                kwargs = {}
                for arg in self.EXPECTED_EVAL_ARGS:
                    kwargs[arg] = batch[arg].to(device)
                output = self.forward(**kwargs)
                probs.append(output.prob.cpu())

            probs = torch.cat(probs).numpy()

        examples = store.data_actor(dsquery)
        cangraphs = store.cangraph_actor(dsquery)
        index = NumpyDataModelHelper.create_simple_index(
            [ex.table.table.table_id for ex in examples],
            [len(c.edgedf) for c in cangraphs],
        )

        output = []
        for ex, cangraph in zip(examples, cangraphs):
            edge_probs = {}
            start, end = index[ex.table.table.table_id]
            for ri, row in enumerate(cangraph.edgedf.iter_rows(named=True)):
                ri = ri + start
                source = str(row["source"])
                target = str(row["target"])
                statement = str(row["statement"])
                inedge = cangraph.edges[row["inedge"]]
                outedge = cangraph.edges[row["outedge"]]

                edge_probs[statement, target, outedge] = probs[ri]
                if outedge == inedge:
                    edge_probs[source, statement, inedge] = probs[ri]
            output.append(edge_probs)
        return DatasetList(examples.name, output)


@dataclass
class PretrainedCTAModelArgs:
    model_class: str
    model_args: dict


class PretrainedCTAModel(BaseActor[PretrainedCTAModelArgs]):
    VERSION = 101

    MOD_SELECTED_TYPE_SCORE = 100

    def __init__(self, params: PretrainedCTAModelArgs, cangraph_actor: CanGraphActor):
        super().__init__(params, [cangraph_actor])
        self.cangraph_actor = cangraph_actor

    @Cache.cache(backend=MemBackend())
    def __call__(self, dsquery: str, verbose: bool) -> DatasetList[NodeProb]:
        return self.method.predict_dataset(
            self.cangraph_actor, dsquery, self.params.model_args, verbose
        )

    @cached_property
    def method(self):
        self.model_class: type[HeuristicELCTAModel] = import_attr(
            self.params.model_class
        )
        return self.model_class()

    def cta_score_offset(self) -> float:
        return self.method.cta_score_offset(self.params.model_args)


class HeuristicELCTAModel:
    def predict_dataset(
        self, actor: CanGraphActor, dsquery: str, args: dict, verbose: bool
    ) -> DatasetList[NodeProb]:
        examples = actor.data_actor(dsquery)
        candidates = actor.canrank_actor.get_candidate_entities(dsquery)

        if args.get("topk", None) is not None:
            candidates = candidates.top_k_candidates(args["topk"])

        ress = predict_cta(
            actor.data_actor.get_kgdb(dsquery),
            examples,
            candidates,
            predict_column_prob,
            {"max_extend_distance": args["max_extend_distance"]},
        )

        output = []
        for res in ress:
            nodeprob = {}
            for ci, ctypes in res.col2type_freq.items():
                nodeprob[ci] = {ctype: cscore for ctype, cscore in ctypes}

                if args["modify_selected_type_score"]:
                    # changing the score of the correct type to be the highest one + 100
                    max_score = max(ctypes, key=itemgetter(1))[1]
                    assert max_score < 100
                    type, score = res.col2type[ci]
                    score = score + 100
                    nodeprob[ci][type] = score
            output.append(nodeprob)
        return DatasetList(examples.name, output)

    def cta_score_offset(self, args: dict) -> float:
        if args["modify_selected_type_score"]:
            return 100
        return 0.0
        return 0.0



# %%
from __future__ import annotations

import re
from functools import lru_cache, partial
from typing import Callable, Optional

import inflect


class NamingHelper:

    @staticmethod
    def norm(name: str, funcs: list[Callable[[str], str]]):
        for f in funcs:
            name = f(name)
        return name

    @staticmethod
    def make_dataset_v2_norm_fn(name: str):
        """The normalization function use in make_dataset_v2"""
        return NamingHelper.norm(
            name,
            funcs=[
                partial(NamingHelper.norm_ref_and_special_chars, tail=False),
                partial(NamingHelper.norm_plurals, parenthesized_plurals=True),
                NamingHelper.norm_spaces,
                partial(NamingHelper.mask_number, whole_word=False),
                str.lower,
            ],
        )

    @staticmethod
    def norm_spaces(name: str):
        words = re.split(r"([^\xa0\n\t\r\f ]+)", name)
        return " ".join((words[i] for i in range(1, len(words), 2)))

    @staticmethod
    def norm_plurals(name: str, parenthesized_plurals: bool = True):
        """Convert plural words in the string to singular form. This function keeps the spaces between words.

        Args:
            name: The input string.
            parenthesized_plurals: If true, remove parenthesized plurals such as `Director(s)`, `Winners(s) and nominee(s)`.
        """
        p = get_inflect_engine()
        # split but keep the separator
        words = re.split(r"([^\xa0\n\t\r\f ]+)", name)
        for i in range(1, len(words), 2):
            # has side effect that pronouns are converted as well (they -> it)
            sing = p.singular_noun(words[i])
            if sing:
                words[i] = sing

            if parenthesized_plurals:
                if words[i].endswith("(s)"):
                    words[i] = words[i][:-3]

        return "".join(words)

    @staticmethod
    def norm_ref_and_special_chars(name: str, tail: bool = True):
        """Detect and remove the following artifacts at anywhere or at end of the strings:

        1. References: [8], [7][8], [a], [b]
        2. Non alphanumeric characters: `:`
        3. Remove (v t e) found in many wikipedia tables
        """
        if tail:
            # we are going to use $ to match the end of the string
            postfix = r"[\xa0\n\t\r\f ]*$"
        else:
            postfix = ""

        is_updated = True

        # remove artifacts that are easy to detect because of special characters such as ( or [
        while is_updated:
            is_updated = False
            for endpattern in [
                r"(\[\d+\])+" + postfix,
                r"(\[[a-zA-Z]\])+" + postfix,
                r"([:;!])" + postfix,
                r"(\n?v\nt\ne)" + postfix,
            ]:
                m = re.search(endpattern, name)
                if m is not None:
                    name = name[: m.span()[0]] + name[m.span()[1] :]
                    is_updated = True

        return name

    @staticmethod
    def mask_number(name: str, whole_word: bool) -> str:
        """Masking numbers in a string to the special character D.

        For example:
            - `2015 Team` -> `DDDD Team`
            - `after 32 events` -> `after DD events`

        Args:
            name: The input string.
            whole_word: If true, only the whole word numbers are masked. Otherwise, every digit is masked.
        """
        if whole_word:
            words = re.split(r"([^\xa0\n\t\r\f ]+)", name)
            assert words[0] == ""
            for i in range(1, len(words), 2):
                if words[i].isdigit():
                    words[i] = "D" * len(words[i])
            return "".join(words)

        return re.sub(r"\d", "D", name)


@lru_cache()
def get_inflect_engine():
    return inflect.engine()


if __name__ == "__main__":
    testcases = [
        "United States (retail)[24]",
        "School [13][15]",
        "Title:",
        "Film![5]",
        "Title\n[19]",
        "Location(s)[3]",
        "Note(s)",
        "Type[3][b]",
        "State[2][3][4]",
        "Districts",
        "Winners and\xa0nominees",
        "Winner(s) and\xa0nominee(s)",
        "after 32 events",
        "after 2012/10 events",
        "Play-by-Play",  # to lowercase is just a side-effect of inflect singular_noun
        "Name\nv\nt\ne",
        "v\nt\ne\nWild Card team\n(Top team qualifies for postseason)",
        "!Name",
    ]

    norm_fns = [
        partial(NamingHelper.norm_ref_and_special_chars, tail=False),
        partial(NamingHelper.norm_plurals, parenthesized_plurals=True),
        NamingHelper.norm_spaces,
        partial(NamingHelper.mask_number, whole_word=False),
    ]

    for s in testcases:
        print(f"`{s}` -> `{NamingHelper.norm(s, norm_fns)}`")

# %%



from __future__ import annotations

from dataclasses import dataclass, field

import gp_core
import numpy as np
from gp.actors.data import KGDB, KGDBArgs
from gp.entity_linking.candidate_generation.common import populate
from gp.misc.conversions import to_rust_table
from loguru import logger
from sm.dataset import Example, FullTable
from sm.misc.ray_helper import enhance_error_info, ray_map, ray_put
from timer import watch_and_report


@dataclass
class LocalSearchArgs:
    similarity: str = field(
        default="levenshtein",
        metadata={
            "help": "augment candidate entities discovered from matching entity properties with values in the same row. "
            "Using the similarity function and threshold to filter out irrelevant candidates. "
        },
    )
    threshold: float = field(
        default=2.0,
        metadata={
            "help": "add candidate entities discovered from matching entity properties with values in the same row. "
            "Any value greater than 1.0 mean we do not apply the similarity function"
        },
    )
    use_column_name: bool = field(
        default=False,
        metadata={
            "help": "add column name to the search text. This is useful for value such as X town, Y town, Z town. "
        },
    )
    search_all_columns: bool = field(
        default=False,
        metadata={"help": "search all columns instead of just the entity columns."},
    )


class LocalSearch:
    VERSION = 101

    def __init__(self, args: LocalSearchArgs, db: KGDB):
        self.args = args
        self.db = db

    def augment_candidates(
        self, examples: list[Example[FullTable]], candidates: DatasetCandidateEntities
    ):
        cfg = gp_core.algorithms.CandidateLocalSearchConfig(
            self.args.similarity,
            self.args.threshold,
            self.args.use_column_name,
            None,
            self.args.search_all_columns,
        )

        logger.debug("prepare localsearch inputs...")
        tables = []
        table_cans = []
        for ex in examples:
            tcans = candidates.get_table_candidates(ex.table.table.table_id)
            table = to_rust_table(ex, tcans)
            tables.append(table)
            table_cans.append(tcans)

        logger.debug(
            "prepare localsearch inputs... done! running candidate local search..."
        )
        newtables = gcore.algorithms.par_candidate_local_search(
            self.db.rudb, tables, None, cfg
        )
        logger.debug("running candidate local search... done!")

        return DatasetCandidateEntities.from_table_candidates(
            {
                ex.table.table.table_id: extract_candidates_from_rust_table(
                    self.db, table_cans[ei], ex.table, newtables[ei]
                )
                for ei, ex in enumerate(examples)
            }
        )

    def augment_candidates_ray(
        self, examples: list[Example[FullTable]], candidates: DatasetCandidateEntities
    ):
        using_ray = len(examples) > 1
        dbref = ray_put(self.db.args, using_ray)
        paramref = ray_put(self.args, using_ray)

        table_cans = ray_map(
            rust_augment_candidates,
            [
                (
                    dbref,
                    ex,
                    candidates.get_table_candidates(ex.table.table.table_id),
                    paramref,
                    not using_ray,
                )
                for ex in examples
            ],
            verbose=True,
            desc="augmenting candidates",
            using_ray=using_ray,
            is_func_remote=False,
            auto_shutdown=True,
        )
        return DatasetCandidateEntities.from_table_candidates(
            {ex.table.table.table_id: tcans for ex, tcans in zip(examples, table_cans)}
        )


@enhance_error_info("1.table.table.table_id")
def rust_augment_candidates(
    kgdb: KGDB | KGDBArgs,
    example: Example[FullTable],
    candidates: TableCandidateEntities,
    params: LocalSearchArgs,
    verbose: bool,
):
    kgdb = KGDB.get_instance(kgdb)
    cfg = gcore.algorithms.CandidateLocalSearchConfig(
        params.similarity,
        params.threshold,
        params.use_column_name,
        None,
        params.search_all_columns,
    )

    newtable = to_rust_table(example, candidates)
    with watch_and_report(
        "create algorithm context",
        preprint=True,
        print_fn=logger.debug,
        disable=not verbose,
    ):
        context = kgdb.rudb.get_algo_context(newtable, n_hop=1, parallel=False)
    with watch_and_report(
        "Performing local search",
        preprint=True,
        print_fn=logger.debug,
        disable=not verbose,
    ):
        newtable = gcore.algorithms.candidate_local_search(
            newtable, context, cfg, False
        )

    return extract_candidates_from_rust_table(kgdb, candidates, example.table, newtable)


def extract_candidates_from_rust_table(
    kgdb: KGDB,
    candidates: TableCandidateEntities,
    pytable: FullTable,
    rutable: gcore.models.Table,
) -> TableCandidateEntities:
    nrows, ncols = pytable.table.shape()
    newcans = {}
    for ci in range(ncols):
        newcans[ci] = {}
        for ri in range(nrows):
            tmp_links = rutable.get_links(ri, ci)
            assert len(tmp_links) <= 1
            if len(tmp_links) == 0:
                assert (
                    not candidates.has_cell_candidates(ri, ci)
                    or len(candidates.get_cell_candidates(ri, ci)) == 0
                )
                continue

            tmp_link = tmp_links[0]
            if not candidates.has_cell_candidates(ri, ci):
                ids = [c.id.id for c in tmp_link.candidates]
                # print("row", ri, "col", ci, "new ids", ids)

                ids, labels, descs, aliases, popularity = populate(kgdb, ids)

                newcans[ci][ri] = CellCandidateEntities(
                    index=None,
                    id=np.array(ids, dtype=np.object_),
                    label=np.array(labels, dtype=np.object_),
                    description=np.array(descs, dtype=np.object_),
                    aliases=np.array(aliases, dtype=np.object_),
                    popularity=np.array(popularity, dtype=np.float64),
                    score=np.array(
                        [c.probability for c in tmp_link.candidates],
                        dtype=np.float64,
                    ),
                    provenance=np.array(["local_search"] * len(ids), dtype=np.object_),
                )
            else:
                cell_cans = candidates.get_cell_candidates(ri, ci)
                existing_ids = set(cell_cans.id)

                new_ids = []
                new_scores = []
                for c in tmp_link.candidates:
                    if c.id.id not in existing_ids:
                        new_ids.append(c.id.id)
                        new_scores.append(c.probability)

                if len(new_ids) > 0:
                    # print("row", ri, "col", ci, "new ids", new_ids)
                    new_ids, labels, descs, aliases, popularity = populate(
                        kgdb, new_ids
                    )
                    newcans[ci][ri] = CellCandidateEntities(
                        index=None,
                        id=np.concatenate(
                            [cell_cans.id, np.array(new_ids, dtype=np.object_)]
                        ),
                        label=np.concatenate(
                            [cell_cans.label, np.array(labels, dtype=np.object_)]
                        ),
                        description=np.concatenate(
                            [
                                cell_cans.description,
                                np.array(descs, dtype=np.object_),
                            ]
                        ),
                        aliases=np.concatenate(
                            [cell_cans.aliases, np.array(aliases, dtype=np.object_)]
                        ),
                        popularity=np.concatenate(
                            [
                                cell_cans.popularity,
                                np.array(popularity, dtype=np.float64),
                            ]
                        ),
                        score=np.concatenate(
                            [cell_cans.score, np.array(new_scores, dtype=np.float64)]
                        ),
                        provenance=np.concatenate(
                            [
                                cell_cans.provenance,
                                np.array(
                                    ["local_search"] * len(new_ids),
                                    dtype=np.object_,
                                ),
                            ]
                        ),
                    )
                else:
                    newcans[ci][ri] = cell_cans

    return TableCandidateEntities.from_cell_candidates(newcans)



"""Generate label data from the table."""

from dataclasses import dataclass
from operator import itemgetter
from typing import Mapping

from gp.distantsupervision.make_dataset.interface import LabelFn
from kgdata.models.entity import Entity
from kgdata.models.ont_class import OntologyClass
from kgdata.wikidata.models.wdclass import WDClass
from kgdata.wikidata.models.wdentity import WDEntity
from sm.dataset import FullTable
from sm.inputs.link import EntityId, EntityIdWithScore, Link
from sm.namespaces.utils import KGName


@dataclass
class LabelV1Args:
    topk: int = 1
    threshold: float = 0.7
    include_similar_score: bool = False


class LabelV1(LabelFn):
    VERSION = 101

    def __init__(
        self,
        args: LabelV1Args,
        entities: Mapping[str, WDEntity | Entity],
        pagerank: Mapping[str, float],
        classes: Mapping[str, OntologyClass],
    ):
        self.args = args
        self.entities = entities
        self.pagerank = pagerank
        self.classes = classes

    def label(
        self, table: FullTable, columns: list[int]
    ) -> list[list[EntityIdWithScore]]:
        return [self.label_column(table.links[:, ci]) for ci in columns]

    def label_column(self, links: list[list[Link]]) -> list[EntityIdWithScore]:
        type2freq = get_type_freq(links, self.entities, self.classes)
        output = []
        for c, freq in sorted(type2freq.items(), key=itemgetter(1), reverse=True):
            if freq < self.args.threshold:
                break

            if len(output) >= self.args.topk and (
                not self.args.include_similar_score
                or (self.args.include_similar_score and output[-1].score != freq)
            ):
                break

            output.append(
                EntityIdWithScore(
                    EntityId(c, KGName.Wikidata),
                    freq,
                )
            )
        return output


def get_type_freq(
    links: list[list[Link]],
    entities: Mapping[str, WDEntity | Entity],
    classes: Mapping[str, OntologyClass],
) -> dict[str, float]:
    """Calculating frequency of types in a column.
    Borrow from gramsplus.algorithms.inferences_v2.features.node_feature.
    Difficult to reuse as it requires data graph/candidate graph.
    """
    n_linked_rows = 0
    type2freq = {}

    for ri in range(len(links)):
        found_classes = set()
        entity_ids = [e for link in links[ri] for e in link.entities]
        for entity_id in entity_ids:
            ent = entities[str(entity_id)]
            for ent_type in ent.instance_of():
                if ent_type not in classes:
                    # sometimes people just tag things incorrectly, e.g.,
                    # Q395 is not instanceof Q41511 (Q41511 is not a class)
                    continue
                found_classes.add(ent_type)

        for c in found_classes:
            if c not in type2freq:
                type2freq[c] = 0
            type2freq[c] += 1

        if len(entity_ids) > 0:
            n_linked_rows += 1

    return {c: freq / n_linked_rows for c, freq in type2freq.items()}


def filter_rows_miss_type(
    table: FullTable, ci: int, types: set[str], entities: Mapping[str, WDEntity]
) -> list[int]:
    nrows = len(table.table.get_column_by_index(ci).values)
    links = table.links

    miss_type_rows = []

    for ri in range(nrows):
        entity_ids = (e for link in links[ri, ci] for e in link.entities)
        if all(
            ent_type not in types
            for entity_id in entity_ids
            for ent_type in entities[str(entity_id)].instance_of()
        ):
            miss_type_rows.append(ri)

    return miss_type_rows



from __future__ import annotations

import enum
import re

import numpy as np
import strsim
from gp.actors.data.prelude import GPExample
from gp.entity_linking.candidate_ranking.feats.feat_fn import CRFeatFn
from gp.entity_linking.candidate_ranking.feats.get_candidates import GetCandidatesFn
from ream.cache_helper import Cache
from ream.data_model_helper._numpy_model import Single2DNumpyArray
from sm.misc.funcs import batch
from sm.misc.ray_helper import is_parallelizable, ray_map

char_tokenizer = strsim.CharacterTokenizer()


CRModelFreeFeats = Single2DNumpyArray


class GetModelFreeFeatFn(CRFeatFn):
    use_args = []

    get_candidates: GetCandidatesFn

    @Cache.cache(
        backend=Cache.sqlite.serde(
            cls=CRModelFreeFeats,
            filename=lambda slf, ex: slf.cache_file + "/data.sqlite",
            mem_persist=True,
        ),
        cache_key=lambda self, ex: ex.id,
        disable="disable",
    )
    def __call__(self, ex: GPExample):
        kgdb_service = self.store.kgdb_service(ex.kgname)

        cans = self.get_candidates(ex)
        labels = kgdb_service.labels(ex.id, cans.ent_id)
        aliases = kgdb_service.aliases(ex.id, cans.ent_id)
        popularities = kgdb_service.pageranks(ex.id, cans.ent_id)

        return CRModelFreeFeats(
            extract_features(cans.cell, labels, aliases, popularities)
        )

    @Cache.flat_cache(
        backend=Cache.sqlite.serde(
            cls=CRModelFreeFeats,
            filename=lambda slf, ex: slf.cache_file + "/data.sqlite",
            mem_persist=True,
        ),
        cache_key=lambda self, ex: ex.id,
        disable="disable",
    )
    def batch_call(self, exs: list[GPExample]) -> list[CRModelFreeFeats]:
        if not is_parallelizable() or len(exs) == 0:
            return [self(ex) for ex in exs]

        kgdb_service = self.store.kgdb_batch_service(exs[0].kgname)

        lst_cans = self.get_candidates.batch_call(exs)
        lst_args = [(ex.id, cans.ent_id) for ex, cans in zip(exs, lst_cans)]
        lst_labels = kgdb_service.batch_labels(lst_args)
        lst_aliases = kgdb_service.batch_aliases(lst_args)
        lst_popularities = kgdb_service.batch_pageranks(lst_args)

        features = ray_map(
            extract_features,
            [
                (
                    lst_cans[ei].cell,
                    lst_labels[ei],
                    lst_aliases[ei],
                    lst_popularities[ei],
                )
                for ei in range(len(exs))
            ],
            verbose=False,
            desc="extract features",
            is_func_remote=False,
        )
        return [CRModelFreeFeats(features[ei]) for ei in range(len(exs))]


def extract_features(
    cells: np.ndarray,
    labels: list[str],
    aliases: list[list[str]],
    popularities: list[float],
):
    feats = []

    for i in range(len(labels)):
        cell = cells[i]

        feat = np.zeros((N_PAIRWISE_FEATURES + N_EXTRA_FEATURES,), dtype=np.float64)
        label_feat = extract_pairwise_features_v2(cell, labels[i])
        for alias in aliases[i]:
            label_feat = np.maximum(
                label_feat, extract_pairwise_features_v2(cell, alias)
            )

        feat[:N_PAIRWISE_FEATURES] = label_feat
        feat[N_PAIRWISE_FEATURES] = popularities[i]
        feats.append(feat)

    return np.array(feats, dtype=np.float64)


# number of extracted pairwise features
# for setting default gold_features for NIL entity
# one extra feature is added from pagerank
class PairwiseFeatures(enum.Enum):
    levenshtein_sim = 0
    jaro_winkler_sim = 1
    monge_elkan_sim = 2
    sym_monge_elkan_sim = 3
    hybrid_jaccard_sim = 4
    numeric_sym_monge_elkan_sim = 5
    numeric_hybrid_jaccard_sim = 6


N_PAIRWISE_FEATURES = len(list(PairwiseFeatures))
N_EXTRA_FEATURES = 1


def extract_pairwise_features_v2(text: str, entity_label: str):
    chartok = strsim.CharacterTokenizer()
    charseqtok = strsim.WhitespaceCharSeqTokenizer()

    text_t1 = chartok.tokenize(text)
    entity_label_t1 = chartok.tokenize(entity_label)

    text_t2 = charseqtok.tokenize(text)
    entity_label_t2 = charseqtok.tokenize(entity_label)

    text_t3 = charseqtok.unique_tokenize(text)
    entity_label_t3 = charseqtok.unique_tokenize(entity_label)

    out2 = [
        strsim.levenshtein_similarity(text_t1, entity_label_t1),
        strsim.jaro_winkler_similarity(text_t1, entity_label_t1),
        strsim.monge_elkan_similarity(text_t2, entity_label_t2),
        (
            sym_monge_score := strsim.symmetric_monge_elkan_similarity(
                text_t2, entity_label_t2
            )
        ),
        (hyjac_score := strsim.hybrid_jaccard_similarity(text_t3, entity_label_t3)),
        does_ordinal_match(text, entity_label, sym_monge_score, 0.7),
        does_ordinal_match(text, entity_label, hyjac_score, 0.7),
    ]
    return out2


def does_ordinal_match(s1: str, s2: str, sim: float, threshold: float) -> float:
    """Test for strings containing ordinal categorical values such as Su-30 vs Su-25, 29th Awards vs 30th Awards.

    Args:
        s1: Cell Label
        s2: Entity Label
    """
    if sim < threshold:
        return 0.4
    digit_tokens_1 = re.findall(r"\d+", s1)
    if len(digit_tokens_1) == 0:
        return 0.4

    digit_tokens_2 = re.findall(r"\d+", s2)
    if digit_tokens_1 == digit_tokens_2:
        return 1.0
    return 0.0



from __future__ import annotations

import re
from operator import attrgetter, itemgetter
from pathlib import Path
from typing import Mapping

import serde.textline
from gramsplus.misc.evaluation.unorganized import IndirectDictAccess, reorder2tree
from kgdata.models.ont_class import OntologyClass
from kgdata.models.ont_property import OntologyProperty


class TaxonomyFn:
    @classmethod
    def normalize_types(
        cls,
        types: list[str] | list[tuple[str, float]],
        predefined_types: dict[str, int],
        collection: Mapping[str, OntologyClass] | Mapping[str, OntologyProperty],
        skip_if_not_found=False,
    ) -> list[str] | list[tuple[str, float]]:
        """Normalize a type to find the closest most-specific one from the taxonomy. If a type
        doesn't have the closest one, return the original type.

        To identify the closest type, related types are sorted by their depth and prefer the one
        appeared in the taxonomy first.
        """
        newtypes = {}
        newtype_scores = {}
        for item in types:
            if isinstance(item, str):
                type, score = item, 0.0
            else:
                type, score = item
            ancestors = collection[type].ancestors
            foundtypes = [
                (pretype, depth)
                for pretype, depth in predefined_types.items()
                if pretype == type or pretype in ancestors
            ]
            if len(foundtypes) > 0:
                foundtype, depth = max(foundtypes, key=itemgetter(1))
                newtypes[foundtype] = max(depth, newtypes.get(foundtype, 0))
                newtype_scores[foundtype] = max(score, newtype_scores.get(foundtype, 0))
            elif not skip_if_not_found:
                newtypes[type] = 100
                newtype_scores[type] = score

        normed_types = [
            item
            for item, _ in sorted(newtypes.items(), key=lambda x: x[1], reverse=True)
        ]
        if isinstance(types[0], str):
            return normed_types

        return [(type, newtype_scores[type]) for type in normed_types]

    @classmethod
    def load_predefined_types(
        cls, high_level_concept_file: Path, classes: Mapping[str, OntologyClass]
    ) -> dict[str, int]:
        # load the predefined high-level types from disk
        predefined_types: dict[str, int] = {}
        tmp = set()
        for line in serde.textline.deser(high_level_concept_file):
            if line.find("#") != -1:
                line = line[: line.find("#")]
            line = line.strip()
            if line == "":
                continue
            tmp.add(cls.label2id(line))

        trees = reorder2tree(
            list(tmp), IndirectDictAccess(classes, attrgetter("ancestors"))
        ).trees
        for tree in trees:
            for item in tree.get_flatten_hierarchy():
                predefined_types[item.id] = max(
                    item.depth, predefined_types.get(item.id, 0)
                )
        assert len(predefined_types) == len(tmp)
        return predefined_types

    @classmethod
    def label2id(cls, label: str):
        m = re.match(r"[^(]*\(([QP]\d+)\)", label)
        assert m is not None, label
        return m.group(1)



from __future__ import annotations

from gp.actors.data.prelude import GPExample
from gp.entity_linking.candidate_ranking.feats.feat_fn import CRFeatFn
from gp.entity_linking.candidate_ranking.feats.get_candidates import CRTableCan, GetCandidatesFn
from gp.entity_linking.candidate_ranking.feats.get_model_free_features import GetModelFreeFeatFn


class GetTopKCandidateMaskFn(CRFeatFn):
    get_candidates: GetCandidatesFn
    get_model_free_feat: GetModelFreeFeatFn

    def __call__(self, ex: GPExample) -> CRTableCan:
        self.get_candidates(ex)



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal, Mapping, Optional

import strsim
from gp.distantsupervision.make_dataset.freg import FilterRegex, FilterRegexArgs
from gp.distantsupervision.make_dataset.utils import FilterMixin, StrPath
from kgdata.models.entity import Entity
from kgdata.wikidata.models.wdentity import WDEntity
from sm.dataset import FullTable
from sm.inputs.link import Link

"""Filtering linked columns in the tables that are noisy:

* Columns that majority of the linked entities are dissimilar to the mentioned.

"""


@dataclass
class FilterV1Args:
    similarity_method: Literal["monge_elkan", "general_jaccard"] = "monge_elkan"
    dissimilarity_threshold: float = 0.2
    noisy_threshold: float = 0.5


class FilterV1(FilterMixin):
    VERSION = 100

    def __init__(
        self,
        args: FilterV1Args,
        entities: Mapping[str, WDEntity | Entity],
        log_file: Optional[StrPath] = None,
    ):
        super().__init__(log_file)
        self.entities = entities
        self.args = args
        self.strsim = StrSim(args.similarity_method)
        self.regex_filter = FilterRegex(FilterRegexArgs(args.noisy_threshold))

    def is_noisy(self, table: FullTable, ci: int) -> tuple[bool, str]:
        if (regex_filter_res := self.regex_filter.is_noisy(table, ci))[0]:
            return regex_filter_res
        values = table.table.get_column_by_index(ci).values

        n_noisy_cells = 0
        for ri, cell in enumerate(values):
            if len(table.links[ri, ci]) == 0:
                continue
            if (
                self.compute_dissimilarity(cell, table.links[ri, ci])
                >= self.args.dissimilarity_threshold
            ):
                n_noisy_cells += 1

        nfreq = n_noisy_cells / len(values)
        return nfreq >= self.args.noisy_threshold, f"{nfreq:.03f} cells are noisy"

    def compute_dissimilarity(self, cell: str, links: list[Link]) -> float:
        entities = self.entities
        score = 0.0
        for link in links:
            for entity in link.entities:
                score = max(
                    score,
                    self.strsim.match(
                        cell, self.iter_entity_name(entities[str(entity)])
                    ),
                )
        return 1 - score

    def iter_entity_name(self, entity: WDEntity | Entity):
        for label in entity.label.lang2value.values():
            yield label
        for aliases in entity.aliases.lang2values.values():
            for alias in aliases:
                yield alias


class StrSim:
    def __init__(self, method: Literal["monge_elkan", "general_jaccard"]):
        self.charseqtok = strsim.WhitespaceCharSeqTokenizer()

        if method == "monge_elkan":
            self.tokenize = self.charseqtok.tokenize
            self.sim_fn = strsim.symmetric_monge_elkan_similarity
        else:
            assert method == "general_jaccard"
            self.tokenize = self.charseqtok.unique_tokenize
            self.sim_fn = strsim.hybrid_jaccard_similarity

    def match(self, query: str, keys) -> float:
        query_tokens = self.tokenize(query)
        return max(self.sim_fn(query_tokens, self.tokenize(key)) for key in keys)



from __future__ import annotations

import os
from dataclasses import dataclass
from functools import cached_property, partial
from pathlib import Path
from typing import IO, Sequence, Union

import numpy as np
import orjson
import ray
import serde.pickle
import strsim
from gp.actors.data import KGDB, KGDBArgs
from gp.entity_linking.candidate_generation.common import CanGenBasicMethod
from hugedict.prelude import RocksDBDict, RocksDBOptions, rocksdb_load
from kgdata.db import pack_int, unpack_int
from kgdata.splitter import split_a_list
from kgdata.wikidata.config import WikidataDirCfg
from kgdata.wikidata.db import WikidataDB
from ream.workspace import ReamWorkspace
from sm.misc.funcs import assert_isinstance
from sm.misc.ray_helper import (
    add_ray_actors,
    get_ray_actors,
    is_parallelizable,
    ray_actor_map,
)
from symspellpy import SymSpell, Verbosity, helpers
from tqdm import tqdm


@dataclass
class FuzzySearchArgs:
    symspell_db_parentdir: Union[Path, str]
    max_dictionary_edit_distance: int = 2
    prefix_length: int = 7
    count_threshold: int = 1
    hard_limit: int = 1000
    sim_score_coeff: float = 0.9


class FuzzySearchImpl:

    def __init__(
        self,
        kgdb: KGDB | KGDBArgs,
        params: FuzzySearchArgs,
    ) -> None:
        if not isinstance(kgdb, KGDB):
            kgdb = KGDB(kgdb)

        WD_DIR = Path(os.environ["WD_DIR"])

        WikidataDirCfg.init(WD_DIR)
        ReamWorkspace.init(os.environ["REAM_DIR"])
        ReamWorkspace.get_instance().register_base_paths(
            DATA_DIR=Path(os.environ["DATA_DIR"])
        )

        pgrnk_stat_file = WikidataDirCfg.get_instance().entity_pagerank / "pagerank.pkl"
        pgrnk_stat = serde.pickle.deser(pgrnk_stat_file)

        self.params = params
        self.symspell = PersistentSymSpell.load(
            Path(self.params.symspell_db_parentdir)
            / f"symspell_v2_d{self.params.max_dictionary_edit_distance}_l{self.params.prefix_length}_t{self.params.count_threshold}",
        )
        self.log_pagerank = (
            float(np.log(pgrnk_stat["min"] + 1e-20)),
            float(np.log(pgrnk_stat["max"]) - np.log(pgrnk_stat["min"] + 1e-20)),
        )
        self.label2ids = assert_isinstance(kgdb.pydb, WikidataDB).label2ids.cache()
        self.chartok = strsim.CharacterTokenizer()

    def get_candidates_by_queries(
        self, queries: list[str]
    ) -> dict[str, list[tuple[str, float]]]:
        out = {}
        for query in queries:
            out[query] = self.query(query)
        return out

    def query(self, query: str) -> list[tuple[str, float]]:
        cans = {}
        query_t1 = self.chartok.tokenize(query)
        for item in self.symspell.lookup(query, Verbosity.ALL):
            label = item.term
            sim_score = strsim.levenshtein_similarity(
                query_t1, self.chartok.tokenize(label)
            )

            for id, pagerank in self.label2ids[label]:
                adjusted_score = sim_score * self.params.sim_score_coeff + (
                    1 - self.params.sim_score_coeff
                ) * self.norm_pagerank(pagerank)
                cans[id] = adjusted_score
        return sorted(cans.items(), key=lambda x: x[1], reverse=True)[
            : self.params.hard_limit
        ]

    def norm_pagerank(self, score):
        return (np.log(score + 1e-20) - self.log_pagerank[0]) / self.log_pagerank[1]


class FuzzySearch(CanGenBasicMethod[FuzzySearchArgs]):

    VERSION = 101

    def __init__(
        self,
        kgdb: KGDB,
        params: FuzzySearchArgs,
        soft_limit: int,
    ) -> None:
        super().__init__(
            kgdb,
            params,
            batch_size=1024,
            soft_limit=soft_limit,
        )
        self.num_actors = 6
        self.actor_ns = f"fuzzy_search_{self.kgdb.args.get_key()}"

    @cached_property
    def impl(self):
        return FuzzySearchImpl(self.kgdb.args, self.params)

    def get_candidates_by_queries(
        self,
        queries: list[str],
    ) -> dict[str, list[tuple[str, float]]]:
        out = {}
        if is_parallelizable():
            # construct an actor
            actors: Sequence["ray.ObjectRef[FuzzySearchImpl]"] = get_ray_actors(
                self.actor_ns
            )
            if len(actors) == 0:
                actors = add_ray_actors(
                    FuzzySearchImpl,
                    (self.kgdb.args, self.params),
                    self.actor_ns,
                    size=self.num_actors,
                )

            batch_size = 64
            resp = ray_actor_map(
                [actor.get_candidates_by_queries.remote for actor in actors],
                [
                    (queries[i : i + batch_size],)
                    for i in range(0, len(queries), batch_size)
                ],
            )
            for item in resp:
                out.update(item)
        else:
            out = self.impl.get_candidates_by_queries(queries)

        return out


class PersistentSymSpell(SymSpell):
    def _load_dictionary_stream(
        self,
        corpus_stream: IO[str],
        term_index: int,
        count_index: int,
        separator: str = " ",
    ) -> bool:
        """Loads multiple dictionary entries from a stream of word/frequency
        count pairs.

        **NOTE**: Merges with any dictionary data already loaded.

        Args:
            corpus_stream: A file object of the dictionary.
            term_index: The column position of the word.
            count_index: The column position of the frequency count.
            separator: Separator characters between term(s) and count.

        Returns:
            ``True`` after file object is loaded.
        """
        for line in tqdm(corpus_stream, total=50000000):
            parts = line.rstrip().split(separator)
            if len(parts) < 2:
                continue
            count = helpers.try_parse_int64(parts[count_index])
            if count is None:
                continue
            key = parts[term_index]
            self.create_dictionary_entry(key, count)
        return True

    def save(self, outdir: Path | str):
        """Save the symspell to a directory."""
        outdir = Path(outdir)
        dbs = PersistentSymSpell._get_databases(
            outdir, readonly=False, without_deletes=True
        )

        for name in ["words", "below_threshold_words", "bigrams"]:
            db = dbs[name]
            for key, value in getattr(self, f"_{name}").items():
                db[key] = value
            db.compact()

        PersistentSymSpell._save_deletes(self._deletes, outdir, verbose=True)

        serde.pickle.ser(
            {
                "max_length": self._max_length,
                # SymSpell settings used to generate the above
                "count_threshold": self._count_threshold,
                "max_dictionary_edit_distance": self._max_dictionary_edit_distance,
                "prefix_length": self._prefix_length,
            },
            outdir / "attrs.pkl",
        )

    @staticmethod
    def load(indir: Path | str):
        """Load the symspell from a directory."""
        indir = Path(indir)
        attrs = serde.pickle.deser(indir / "attrs.pkl")
        symspell = PersistentSymSpell(
            max_dictionary_edit_distance=attrs["max_dictionary_edit_distance"],
            prefix_length=attrs["prefix_length"],
            count_threshold=attrs["count_threshold"],
        )

        dbs = PersistentSymSpell._get_databases(indir, readonly=True)

        symspell._count_threshold = attrs["count_threshold"]
        symspell._max_dictionary_edit_distance = attrs["max_dictionary_edit_distance"]
        symspell._prefix_length = attrs["prefix_length"]
        symspell._max_length = attrs["max_length"]
        for name, db in dbs.items():
            setattr(symspell, f"_{name}", db)
        return symspell

    @staticmethod
    def _save_deletes(
        deletes: dict[str, list[str]], outdir: Path, verbose: bool = False
    ):
        dbdir = outdir / "deletes.db"
        if (dbdir / "_SUCCESS").exists():
            return

        rawdir = outdir / "deletes"
        rawdir.mkdir(parents=True, exist_ok=True)
        if not (rawdir / "_SUCCESS").exists():
            kvs = [
                orjson.dumps((key, orjson.dumps(value).decode()))
                for key, value in tqdm(
                    deletes.items(), disable=not verbose, mininterval=1
                )
            ]
            split_a_list(kvs, rawdir / "part.jl")
            (rawdir / "_SUCCESS").touch()

        rocksdb_load(
            dbpath=str(dbdir),
            dbopts=RocksDBOptions(
                create_if_missing=True,
                compression_type="lz4",
                bottommost_compression_type="zstd",
            ),
            files=[str(x) for x in rawdir.glob("part-*.jl")],
            format={
                "record_type": {"type": "tuple2", "key": None, "value": None},
                "is_sorted": False,
            },
            verbose=True,
            compact=True,
        )
        (dbdir / "_SUCCESS").touch()

    @staticmethod
    def _get_databases(outdir: Path, readonly: bool, without_deletes: bool = False):
        words: RocksDBDict[str, int] = RocksDBDict(
            path=str(outdir / "words.db"),
            options=RocksDBOptions(
                create_if_missing=True,
            ),
            deser_key=partial(str, encoding="utf-8"),
            deser_value=unpack_int,
            ser_value=pack_int,
            readonly=readonly,
        )
        below_threshold_words: RocksDBDict[str, int] = RocksDBDict(
            path=str(outdir / "below_threshold_words.db"),
            options=RocksDBOptions(
                create_if_missing=True,
            ),
            deser_key=partial(str, encoding="utf-8"),
            deser_value=unpack_int,
            ser_value=pack_int,
            readonly=readonly,
        )
        bigrams: RocksDBDict[str, int] = RocksDBDict(
            path=str(outdir / "bigrams.db"),
            options=RocksDBOptions(
                create_if_missing=True,
            ),
            deser_key=partial(str, encoding="utf-8"),
            deser_value=unpack_int,
            ser_value=pack_int,
            readonly=readonly,
        )
        out: dict = {
            "words": words,
            "below_threshold_words": below_threshold_words,
            "bigrams": bigrams,
        }

        if not without_deletes:
            deletes: RocksDBDict[str, list[str]] = RocksDBDict(
                path=str(outdir / "deletes.db"),
                options=RocksDBOptions(
                    create_if_missing=True,
                    compression_type="lz4",
                    bottommost_compression_type="zstd",
                ),
                deser_key=partial(str, encoding="utf-8"),
                deser_value=orjson.loads,
                ser_value=orjson.dumps,
                readonly=readonly,
            )
            out["deletes"] = deletes

        return out



from __future__ import annotations

from collections import defaultdict
from copy import deepcopy
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import Literal, Optional

import orjson
from gp.actors.data._db import KGDB, KGDBArgs
from gp.distantsupervision.make_dataset.prelude import (
    CombinedFilter,
    CombinedFilterArgs,
    EntityRecognitionV1,
    EntityRecognitionV1Args,
    FilterByEntType,
    FilterByEntTypeArgs,
    FilterByHeaderColType,
    FilterByHeaderColTypeArgs,
    FilterFn,
    FilterNotEntCol,
    FilterRegex,
    FilterRegexArgs,
    FilterV1,
    FilterV1Args,
    LabelFn,
    LabelV1,
    LabelV1Args,
    LabelV2,
    LabelV2Args,
    NoFilter,
    NoTransform,
    TransformFn,
    TransformV1,
    TransformV1Args,
    TransformV2,
)
from gp.semanticmodeling.text_parser import TextParser
from libactor.cache import BackendFactory, IdentObj, cache
from libactor.misc import get_parallel_executor, orjson_dumps, typed_delayed
from ream.params_helper import NoParams
from sm.dataset import Example, FullTable
from sm.inputs.prelude import EntityIdWithScore
from sm.misc.funcs import batch


@dataclass
class AutoLabelDataActorArgs:
    skip_non_unique_mention: bool = field(
        default=False,
        metadata={
            "help": "Skip tables with non-unique mention",
        },
    )
    skip_column_with_no_type: bool = field(
        default=False,
        metadata={
            "help": "Column that auto-labeler cannot find any entity type will be skipped"
        },
    )
    recog_method: Literal["recog_v1"] = field(
        default="recog_v1",
        metadata={
            "help": "Entity recognition method to use",
            "variants": {"recog_v1": EntityRecognitionV1},
        },
    )
    recog_v1: EntityRecognitionV1Args = field(
        default_factory=EntityRecognitionV1Args,
        metadata={"help": "Entity recognition v1 arguments"},
    )
    filter_method: Literal[
        "no_filter",
        "filter_non_ent_col",
        "filter_regex",
        "filter_v1",
        "filter_ent_type",
        "filter_header_col_type",
        "filter_combined",
    ] = field(
        default="filter_regex",
        metadata={
            "help": "Filter method to use",
            "variants": {
                "no_filter": NoFilter,
                "filter_non_ent_col": FilterNotEntCol,
                "filter_regex": FilterRegex,
                "filter_v1": FilterV1,
                "filter_ent_type": FilterByEntType,
                "filter_header_col_type": FilterByHeaderColType,
                "filter_combined": CombinedFilter,
            },
        },
    )
    filter_regex: FilterRegexArgs = field(
        default_factory=FilterRegexArgs,
        metadata={"help": "Filter regex arguments"},
    )
    filter_ent_type: FilterByEntTypeArgs = field(
        default_factory=FilterByEntTypeArgs,
        metadata={"help": "Filter by entity type arguments"},
    )
    filter_not_ent_col: NoParams = field(
        default_factory=NoParams,
        metadata={"help": "Filter not entity column arguments"},
    )
    filter_header_col_type: Optional[FilterByHeaderColTypeArgs] = field(
        default=None,
        metadata={"help": "Filter by header col type arguments"},
    )
    filter_combined: CombinedFilterArgs = field(
        default_factory=CombinedFilterArgs,
        metadata={"help": "Combined filter arguments"},
    )
    filter_v1: FilterV1Args = field(
        default_factory=FilterV1Args,
        metadata={"help": "Filter v1 arguments"},
    )
    transform_method: Literal["transform_v1", "transform_v2", "no_transform"] = field(
        default="transform_v1",
        metadata={
            "help": "Transformation method to use",
            "variants": {
                "transform_v1": TransformV1,
                "no_transform": NoTransform,
                "transform_v2": TransformV2,
            },
        },
    )
    transform_v1: Optional[TransformV1Args] = field(
        default=None,
        metadata={"help": "Transformation v1 arguments"},
    )
    transform_v2: Optional[TransformV1Args] = field(
        default=None,
        metadata={"help": "Transformation v2 arguments"},
    )
    label_method: Literal["label_v1", "label_v2"] = field(
        default="label_v1",
        metadata={
            "help": "Label method to use",
            "variants": {"label_v1": LabelV1, "label_v2": LabelV2},
        },
    )
    label_v1: Optional[LabelV1Args] = None
    label_v2: Optional[LabelV2Args] = None

    def get_key(self):
        return orjson_dumps(asdict(self)).decode()


@dataclass
class AutoLabeledTable:
    table: FullTable
    entity_columns: list[int]
    entity_column_types: list[list[EntityIdWithScore]]

    def to_dict(self):
        return {
            "table": self.table.to_dict(),
            "entity_columns": self.entity_columns,
            "entity_column_types": [
                [e.to_dict() for e in coltypes] for coltypes in self.entity_column_types
            ],
        }

    @classmethod
    def from_dict(cls, obj: dict):
        return cls(
            table=FullTable.from_dict(obj["table"]),
            entity_columns=obj["entity_columns"],
            entity_column_types=[
                [EntityIdWithScore.from_dict(e) for e in coltypes]
                for coltypes in obj["entity_column_types"]
            ],
        )


def make_autolabel(
    exs: list[Example[FullTable]],
    kgdb: KGDB,
    args: AutoLabelDataActorArgs,
    workdir: Path,
    n_jobs: int = -1,
) -> list[AutoLabeledTable]:
    batch_exs = batch(64, exs)
    return [
        x
        for arr in get_parallel_executor(n_jobs=n_jobs, return_as="generator")(
            typed_delayed(wrap_process_example)(lst, kgdb.args, args, workdir)
            for lst in batch_exs
        )
        for x in arr
    ]


def wrap_process_example(
    exs: list[Example[FullTable]],
    kgdb_args: KGDBArgs,
    args: AutoLabelDataActorArgs,
    workdir: Path,
) -> list[AutoLabeledTable]:
    kgdb = KGDB.get_instance(kgdb_args)
    ikgdb = IdentObj(kgdb.args.get_key(), kgdb)
    iargs = IdentObj(args.get_key(), args)

    er = get_er(iargs)
    filter = get_filter(iargs, ikgdb, workdir)
    map = get_transformation(iargs, ikgdb, workdir)
    labeler = get_label(iargs, ikgdb)
    text_parser = TextParser.default()

    return [
        result
        for ex in exs
        if (result := process_example(ex, er, filter, map, labeler, text_parser, args))
        is not None
    ]


def process_example(
    ex: Example[FullTable],
    er: EntityRecognitionV1,
    fil: FilterFn,
    map: TransformFn,
    labeler: LabelFn,
    text_parser: TextParser,
    args: AutoLabelDataActorArgs,
) -> Optional[AutoLabeledTable]:
    table = ex.table
    if args.skip_non_unique_mention and has_non_unique_mention(table):
        return None

    # recognize entity columns
    entity_columns = er.recognize(table)
    entity_columns = fil.filter(table, entity_columns)

    table = map.transform(table, entity_columns)
    entity_column_types = labeler.label(table, entity_columns)

    if args.skip_column_with_no_type:
        valid_cols = [
            i for i in range(len(entity_columns)) if len(entity_column_types[i]) > 0
        ]
        entity_columns = [entity_columns[i] for i in valid_cols]
        entity_column_types = [entity_column_types[i] for i in valid_cols]

    if len(entity_columns) == 0:
        return None

    newtable = normalize_table(ex.table, text_parser)
    return AutoLabeledTable(newtable, entity_columns, entity_column_types)


def has_non_unique_mention(table: FullTable) -> bool:
    """Check if the example table has the same mention at different cells linked to different entities"""
    col2mention = defaultdict(lambda: defaultdict(set))

    for ri, ci, links in table.links.enumerate_flat_iter():
        if len(links) == 0:
            continue

        text = table.table[ri, ci]
        assert isinstance(text, str), text

        for link in links:
            mention = text[link.start : link.end]
            if len(mention) > 0:
                col2mention[ci][mention].update(link.entities)
                if len(col2mention[ci][mention]) > 1:
                    return True

    return False


def normalize_table(oldtable: FullTable, text_parser: TextParser) -> FullTable:
    table = deepcopy(oldtable)
    for col in table.table.columns:
        assert col.name is not None
        col.name = text_parser._norm_string(col.name)

    # normalize cells and links
    for ci, col in enumerate(table.table.columns):
        for ri, cell in enumerate(col.values):
            if isinstance(cell, str):
                newcell = text_parser._norm_string(cell)
                col.values[ri] = newcell

                if newcell != cell:
                    # adjust the links
                    for link in table.links[ri, ci]:
                        if (
                            cell[link.start : link.end]
                            != newcell[link.start : link.end]
                        ):
                            # the new mention is different from the old mention
                            before = text_parser._norm_nostrip_string(
                                cell[: link.start]
                            ).lstrip()
                            mention = text_parser._norm_nostrip_string(
                                cell[link.start : link.end]
                            )
                            if len(before) == 0 and mention.lstrip() != mention:
                                mention = mention.lstrip()
                            after = text_parser._norm_nostrip_string(
                                cell[link.end :]
                            ).rstrip()
                            if len(after) == 0 and mention.rstrip() != mention:
                                mention = mention.rstrip()
                            if before + mention + after != newcell:
                                raise NotImplementedError(
                                    f"Haven't implemented fixing where part of the mention has been changed. Recovered string: `{before+mention+after}` - transformed string: `{newcell}`"
                                )
                            link.start = len(before)
                            link.end = len(before) + len(mention)
    return table


@cache(backend=BackendFactory.func.mem)
def get_er(args: IdentObj[AutoLabelDataActorArgs]):
    if args.value.recog_method == "recog_v1":
        return EntityRecognitionV1(args.value.recog_v1)
    raise NotImplementedError()


@cache(backend=BackendFactory.func.mem)
def get_filter(
    args: IdentObj[AutoLabelDataActorArgs], kgdb: IdentObj[KGDB], workdir: Path
):
    logfile = workdir / "filter.log"

    vargs = args.value
    vkgdb = kgdb.value

    if vargs.filter_method == "filter_regex":
        return FilterRegex(vargs.filter_regex, logfile)
    if vargs.filter_method == "filter_v1":
        return FilterV1(
            vargs.filter_v1,
            vkgdb.pydb.entities.cache(),
            logfile,
        )
    if vargs.filter_method == "filter_combined":
        filters = [
            FilterRegex(vargs.filter_combined.regex, logfile),
            FilterByEntType(
                vargs.filter_combined.ignore_types,
                vkgdb.pydb.entities.cache(),
                vkgdb.pydb.classes.cache(),
                logfile,
            ),
        ]
        if vargs.filter_combined.header_col_type is not None:
            filters.append(
                FilterByHeaderColType(
                    vargs.filter_combined.header_col_type,
                    get_label(args, kgdb),
                    logfile,
                )
            )
        return CombinedFilter(
            filters,
            logfile,
        )
    if vargs.filter_method == "no_filter":
        return NoFilter()
    if vargs.filter_method == "filter_non_ent_col":
        return FilterNotEntCol(logfile)
    raise NotImplementedError()


@cache(backend=BackendFactory.func.mem)
def get_label(args: IdentObj[AutoLabelDataActorArgs], kgdb: IdentObj[KGDB]) -> LabelFn:
    vargs = args.value
    vkgdb = kgdb.value
    if vargs.label_method == "label_v1":
        assert vargs.label_v1 is not None
        return LabelV1(
            vargs.label_v1,
            vkgdb.pydb.entities.cache(),
            vkgdb.pydb.entity_pagerank.cache(),
            vkgdb.pydb.classes.cache(),
        )
    if vargs.label_method == "label_v2":
        assert vargs.label_v2 is not None
        return LabelV2(
            vargs.label_v2,
            vkgdb.pydb.entities.cache(),
            vkgdb.pydb.entity_pagerank.cache(),
            vkgdb.pydb.classes.cache(),
        )
    raise NotImplementedError()


@cache(backend=BackendFactory.func.mem)
def get_transformation(
    args: IdentObj[AutoLabelDataActorArgs], kgdb: IdentObj[KGDB], workdir: Path
):
    vargs = args.value
    vkgdb = kgdb.value

    logfile = workdir / f"transform.log"
    if vargs.transform_method == "transform_v1":
        assert vargs.transform_v1 is not None
        return TransformV1(
            vargs.transform_v1,
            vkgdb.pydb.entities.cache(),
            vkgdb.pydb.classes.cache(),
            logfile,
        )
    if vargs.transform_method == "transform_v2":
        assert vargs.transform_v2 is not None
        return TransformV2(
            vargs.transform_v2,
            vkgdb.pydb.entities.cache(),
            vkgdb.pydb.classes.cache(),
            logfile,
        )

    if vargs.transform_method == "no_transform":
        return NoTransform()
    raise NotImplementedError()



from __future__ import annotations

from functools import lru_cache
from typing import Callable, TypeVar

import numpy as np
from gp.actors.data.prelude import GPExample
from gp.actors.el.cangen import CanGenActor
from gp.entity_linking.candidate_ranking.feats.feat_fn import CRFeatFn, CRFeatFnArgs
from gp.entity_linking.candidate_ranking.feats.get_candidates import CRTableCan, GetCandidatesFn
from gp.entity_linking.candidate_ranking.feats.get_entity_desc_embedding import (
    GetEmptyEntityDescEmbeddingFn,
)
from gp.entity_linking.candidate_ranking.feats.get_freq_match import GetFreqMatchDataFn
from gp.entity_linking.candidate_ranking.feats.get_header_embedding import (
    GetEmptyHeaderEmbeddingFn,
)
from gp.entity_linking.candidate_ranking.feats.get_model_free_features import GetModelFreeFeatFn
from gp.entity_linking.populate import KGDBBatchService, KGDBService
from gp.misc.dataset import ColumnarDataset, extended_collate_fn
from ream.cache_helper import Cache, MemBackend
from ream.params_helper import NoParams
from ream.prelude import BaseActor
from sm.misc.funcs import batch
from sm.namespaces.utils import KGName
from tqdm import tqdm

Fn = TypeVar("Fn", bound=CRFeatFn)


class CRDatasetBuilder(BaseActor[NoParams]):
    VERSION = 100

    def __init__(
        self,
        cangen_actor: CanGenActor,
    ):
        super().__init__(NoParams(), [cangen_actor])

        self.cangen_actor = cangen_actor

    def __call__(
        self,
        args: CRFeatFnArgs,
        exs: list[GPExample],
        enable_fns: list[type[CRFeatFn]],
        verbose: bool = False,
    ):
        with CRFeatFn.auto_clear_mem_backend():
            funcs: list[tuple[str, CRFeatFn]] = []
            for enable_fn in enable_fns:
                funcs.append(
                    (
                        self.get_func_name(enable_fn),
                        self.get_func(enable_fn).set_args(args),
                    )
                )

            output = {fname: [] for fname, func in funcs}
            for batch_exs in tqdm(batch(64, exs), disable=not verbose):
                for fname, func in funcs:
                    output[fname].extend(func.batch_call(batch_exs))

            return output

    @lru_cache(maxsize=None)
    def get_func(self, fncls: type[Fn]) -> Fn:
        return fncls(self)

    @lru_cache(maxsize=None)
    def get_func_name(self, fncls: type[Fn]) -> str:
        return fncls.__name__

    @lru_cache(maxsize=1)
    def kgdb_service(self, kgname: KGName) -> KGDBService:
        return KGDBService(self.cangen_actor.db_actor.kgdbs[kgname])

    @lru_cache(maxsize=1)
    def kgdb_batch_service(self, kgname: KGName) -> KGDBBatchService:
        return KGDBBatchService(self.cangen_actor.db_actor.kgdbs[kgname].args)


def make_pairwise_v3_nocontext_dataset(
    store: CRDatasetBuilder,
    exs: list[GPExample],
    verbose: bool = False,
):
    out = store(
        CRFeatFnArgs(text_embedding_model="sentence-transformers/all-mpnet-base-v2"),
        exs,
        [
            GetCandidatesFn,
            GetModelFreeFeatFn,
            GetFreqMatchDataFn,
            GetEmptyHeaderEmbeddingFn,
            GetEmptyEntityDescEmbeddingFn,
        ],
        verbose=verbose,
    )

    lst_cans: list[CRTableCan] = out[store.get_func_name(GetCandidatesFn)]
    size = sum(x.cell.shape[0] for x in lst_cans)

    example_ranges = []
    cell_id = np.zeros((size,), dtype=np.int32)
    cell_id_counter = 0
    offset = 0
    for ei, ex in enumerate(exs):
        nrows, ncols = ex.table.table.shape()
        cans = lst_cans[ei]

        cell_index = cell_id_counter + cans.row_index * ncols + cans.col_index
        cell_id[offset : offset + cell_index.shape[0]] = cell_index
        example_ranges.append((offset, offset + cell_index.shape[0]))

        cell_id_counter += nrows * ncols
        offset += cell_index.shape[0]

    return ColumnarDataset(
        {
            "cell_id": cell_id,
            "cell_label": np.concatenate([x.cell for x in lst_cans]),
            "label": np.concatenate([x.is_correct for x in lst_cans]),
            "features": np.concatenate(
                [
                    np.concatenate(
                        [x.value for x in out[store.get_func_name(GetModelFreeFeatFn)]]
                    ),
                    np.concatenate(
                        [x.value for x in out[store.get_func_name(GetFreqMatchDataFn)]]
                    ).reshape(-1, 1),
                ],
                axis=1,
                dtype=np.float32,
            ),
            "header_embedding": np.concatenate(
                out[store.get_func_name(GetEmptyHeaderEmbeddingFn)]
            ),
            "entity_desc_embedding": np.concatenate(
                out[store.get_func_name(GetEmptyEntityDescEmbeddingFn)]
            ),
        },
        collate_fn=extended_collate_fn,
        references={
            "example_ranges": example_ranges,
        },
    )



from __future__ import annotations

from typing import Optional

from gp.distantsupervision.make_dataset.utils import FilterMixin, StrPath
from gp.entity_linking.candidate_recognition._heuristic_model import HeuristicCanReg
from kgdata.models import Ontology
from sm.dataset import Example, FullTable
from sm.misc.fn_cache import CacheMethod


class FilterNotEntCol(FilterMixin):
    VERSION = 100

    def __init__(self, log_file: Optional[StrPath] = None):
        super().__init__(log_file)
        self.canreg = HeuristicCanReg()

    def is_noisy(self, table: FullTable, ci: int) -> tuple[bool, str]:
        if ci not in self.reg_ent_cols(table):
            return True, "The column is not recognized as entity column"
        return False, ""

    @CacheMethod.cache(CacheMethod.single_object_arg)
    def reg_ent_cols(self, table: FullTable) -> set[int]:
        # set the ontology to None because HeuristicCanReg does not use it
        # and we don't want to pass it to the function
        ontology: Ontology = None  # type: ignore
        return set(
            self.canreg(
                [Example(id=table.table.table_id, sms=[], table=table)], ontology
            )[0]
        )



"""
Attempt to fix incorrect auto links in the Wikipedia tables:

1. Detect linking to temporal entity and updating it.
"""

from __future__ import annotations

import enum
import re
from dataclasses import dataclass, field
from typing import Literal, Mapping, Optional

from gp.distantsupervision.make_dataset.interface import TransformFn
from gp.distantsupervision.make_dataset.utils import LoggerMixin, StrPath
from kgdata.models.entity import Entity
from kgdata.models.ont_class import OntologyClass
from kgdata.wikidata.models.wdclass import WDClass
from kgdata.wikidata.models.wdentity import WDEntity
from loguru import logger
from sm.dataset import FullTable
from sm.inputs.link import EntityId, Link
from sm.misc.funcs import assert_not_null


class ChangeStatus(enum.IntEnum):
    NoChange = 0
    Update = 1
    Remove = 2


DEFAULT_TEMPORAL_CLASSES = {
    "Q46135307",  # nation at sport competition
    "Q27020041",  # sports season
    "Q29791211",  # sport in a geographic region
}
DEFAULT_SURGERIES = [
    "Q46135307:P17",  # nation at sport competition -> country via P17 (country)
    "Q46135307:P1532",  # nation at sport competition -> country for sport (P1532) -- noisy
    "Q1539532:P5138",  # sports season of a sports club -> club via P5138 (season of club or team)
    "Q27020041:P3450",  # sports season (Q27020041) -> sports season of league or competition (P3450) -- noisy
    "Q29791211:P17",  # sport in a geographic region (Q29791211) -> country (P17)
]


@dataclass
class TransformV1Args:
    noisy_threshold: float = field(
        default=0.5,
        metadata={
            "help": "When the freq of noisy cells in the column exceed the threshold, the transformation on that column is triggered."
        },
    )
    temporal_classes: set[str] = field(
        default_factory=lambda: DEFAULT_TEMPORAL_CLASSES,
        metadata={
            "help": "Temporal classes that are used to detect temporal entities."
        },
    )
    known_surgeries: list[str] = field(
        default_factory=lambda: DEFAULT_SURGERIES,
        metadata={
            "help": "Known surgeries of <classid>:[<prop>] or <classid>:<prop> where [] denote optional"
        },
    )


class TransformV1(TransformFn, LoggerMixin):
    VERSION = 104

    def __init__(
        self,
        args: TransformV1Args,
        entities: Mapping[str, WDEntity | Entity],
        classes: Mapping[str, OntologyClass],
        log_file: Optional[StrPath] = None,
    ):
        super().__init__(log_file)
        self.args = args
        self.entities = entities
        self.classes = classes
        self.temporal_classes = args.temporal_classes

        regex = re.compile(r"(\w+):(\w+)")
        self.known_surgeries: list[Surgery] = [
            Surgery(
                steps=[
                    SurgeryStep(
                        (m := assert_not_null(regex.match(sur))).group(1),
                        m.group(2),
                    )
                    for sur in surpath.split(" -> ")
                ],
                optional=surpath.startswith("[") and surpath.endswith("]"),
            )
            for surpath in args.known_surgeries
        ]

    def transform(self, table: FullTable, columns: list[int]) -> FullTable:
        nrows, ncols = table.table.shape()
        newlinks = None
        for ci in columns:
            collinks = self.transform_column(table, ci)
            if collinks is not None:
                if newlinks is None:
                    newlinks = table.links.shallow_copy()
                for ri in range(nrows):
                    newlinks[ri, ci] = collinks[ri]
        if newlinks is None:
            return table

        self.logger.debug("Transformed table {}", table.table.table_id)
        return FullTable(table.table, table.context, newlinks)

    def transform_column(
        self, table: FullTable, column: int
    ) -> Optional[list[list[Link]]]:
        nrows, ncols = table.table.shape()
        n_transformed = 0
        collinks = []
        for ri in range(nrows):
            newlinks = self.transform_cell(table, ri, column)
            if newlinks is not None:
                collinks.append(newlinks)
                n_transformed += 1
            else:
                collinks.append(table.links[ri, column])
        if n_transformed / nrows >= self.args.noisy_threshold:
            return collinks
        return None

    def transform_cell(
        self, table: FullTable, row: int, column: int
    ) -> Optional[list[Link]]:
        """Transform a cell if possible:

        1. empty mentions are ignored even if they are linked (e.g., country flag).
        2. entity that are links to temporal entities are replaced with non-temporal entity.
        """
        non_empty_links = [
            link for link in table.links[row, column] if link.end > link.start
        ]

        modified = len(non_empty_links) != len(table.links[row, column])
        newlinks = []
        for link in non_empty_links:
            newents = []
            for entid in link.entities:
                ent = self.entities[entid]
                newentid, status = self.transform_entity(ent)
                modified = modified or status != ChangeStatus.NoChange
                if status == ChangeStatus.NoChange:
                    newents.append(entid)
                elif status == ChangeStatus.Update:
                    newents.append(EntityId(newentid, entid.type))
                elif status == ChangeStatus.Remove:
                    continue
            if modified:
                newlinks.append(
                    Link(
                        start=link.start,
                        end=link.end,
                        url=link.url,
                        entities=newents,
                    )
                )

        if modified:
            return newlinks
        return None

    def transform_entity(self, entity: WDEntity | Entity) -> tuple[str, ChangeStatus]:
        """Transform an entity if possible.

        It detects if the entity is an instance of temporal classes such as Football Season, and if found, it will resolve to the correct entity using predefined rules that
        select the value of the property.
        """
        entclasses = [self.classes[cid] for cid in entity.instance_of()]
        if not any(
            any(
                tempcls == entclass.id or tempcls in entclass.ancestors
                for tempcls in self.temporal_classes
            )
            for entclass in entclasses
        ):
            return entity.id, ChangeStatus.NoChange

        found_surgeries = False
        for surgery in self.known_surgeries:
            newentid, stt = self.try_surgery(surgery, entity, entclasses)
            if stt == "notapplicable":
                continue

            found_surgeries = True

            if stt == "success":
                return newentid, ChangeStatus.Update

            if stt == "nodata":
                # this will print a warning, and we continue to try if there is any other surgery
                # that can be applied. but if there is no surgery that can be applied, we will
                # remove the entity.
                continue

        if not found_surgeries:
            raise Exception(f"Unknown surgery {entity.label} ({entity.id})")
        return entity.id, ChangeStatus.Remove

    def try_surgery(
        self,
        surgery: Surgery,
        entity: WDEntity | Entity,
        entclasses: list[OntologyClass],
    ) -> tuple[str, Literal["nodata", "notapplicable", "success"]]:
        current_ent = entity
        current_entclasses = entclasses

        n = len(surgery.steps)

        for i, step in enumerate(surgery.steps):
            if not any(
                step.type == entclass.id or step.type in entclass.ancestors
                for entclass in current_entclasses
            ):
                if i > 0 and not surgery.optional:
                    self.logger.warning(
                        f"The next step requires type {step.type}, but entity {current_ent.label} ({current_ent.id}) isn't of that type."
                    )
                    return "", "nodata"
                return "", "notapplicable"
            if step.prop not in current_ent.props:
                if not surgery.optional:
                    self.logger.warning(
                        f"Missing {step.prop} for {current_ent.label} ({current_ent.id})"
                    )
                return "", "nodata"

            potential_values = set(current_ent.get_object_prop_value(step.prop))
            if len(potential_values) > 1:
                # there are too many values, we don't know which one to choose.
                if entity.id == "Q2366374":
                    # hard code a few cases as it's take time more time to write code to handle them.
                    # these cases are error in Wikidata.
                    potential_values = {"Q213347"}
                elif entity.id == "Q749109" and step.prop == "P17":
                    potential_values = {"Q145"}
                else:
                    raise Exception(
                        f"Too many values for {current_ent.label} ({current_ent.id}) -> {step.prop}"
                    )

            if i == n - 1:
                return next(iter(potential_values)), "success"

            current_ent = self.entities[next(iter(potential_values))]
            current_entclasses = [
                self.classes[cid] for cid in current_ent.instance_of()
            ]

        raise Exception("Shouldn't reach here")


@dataclass
class Surgery:
    steps: list[SurgeryStep]
    optional: bool


@dataclass
class SurgeryStep:
    type: str
    prop: str
